// Copyright (C) 2008 Cory Strope <corystrope@gmail.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdio.h>
#include <stdlib.h>
#include <cstring>
#include <math.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <string>
#include <iomanip>

#include "aamodels.h"
#include "insert_freqs.h"
#include "tree.h"
#include "evolve.h"
#include "model.h"
#include "gamma.h"
#include "twister.h"
#include "treefile.h"
#include "motif.h"

#define INSERT 0
#define DELETE 1

#define FSL_FAIL -1
#define PRINT_ID 0

using namespace std;

int		node_num = 0;
int		total_events = 0;
static double NeighborRates[400];
int indel_fill_aas;
double freqRate[MAX_RATE_CATS];
static double *matrix[MAX_RATE_CATS];
static double *cvector;
int indelNo, num_insert, num_delete, len_insert, len_delete;
int num_subst = 0;
bool insert_debug = false;
bool delete_debug = false;

// functions 

void CreateMatrix() 
{
    for (int i = 0; i < MAX_RATE_CATS; i++) {
		try {
			matrix[i] = new double [numStates * numStates];
		} catch (exception& e) {
			cerr << "Error allocating matrix: " << e.what() << endl;
			exit(EXIT_FAILURE);
		}
    }
	if (!isNucModel) Set_Neighbor_Rates(NeighborRates,indel_fill_aas);
	try {
		cvector = new double [numStates];
	} catch (exception& e) {
		cerr << "Error allocating cvector: " << e.what() << endl;
		exit(EXIT_FAILURE);
	}
}

void DestroyMatrix() 
{
    for (int i = 0; i < MAX_RATE_CATS; i++) {
		try {
			delete [] matrix[i];
		} catch (exception& e) {
			cerr << "Error de-allocating matrix: " << e.what() << endl;
			exit(EXIT_FAILURE);
		}
    }
	try {
		delete [] cvector;
	} catch (exception& e) {
		cerr << "Error de-allocating cvector: " << e.what() << endl;
		exit(EXIT_FAILURE);
	}
}

void CreateRates(TNode *node, int inNumSites) 
{
    if (node->nodeEnv->rateHetero==GammaRates) {
		try {
    		node->gammaRates.assign(inNumSites, 0);
		} catch (exception &e) {
			cerr << "Could not assign gammaRates for node ";
			node->report();
			cerr << ": " << e.what() << endl;
			exit(EXIT_FAILURE);
		}
    } else if (node->nodeEnv->rateHetero==DiscreteGammaRates) {
		try {
   	 		node->categories.assign(inNumSites, 0);
		} catch (exception& e) {
			cerr << "Could not assign categories for node ";
			node->report();
			cerr << ": " << e.what() << endl;
			exit(EXIT_FAILURE);		
		}
	}
}

void Create_invarRates(TNode *node, int inNumSites) 
{
	if (node->nodeEnv->invariableSites) node->invariable.assign(inNumSites, 0);
//    else cerr << "Create_invarRates: invariable array NOT created." << endl;
}

void Create_Global_Arrays(TTree *tree, int inNumSites) 
{
	vector<char>::iterator cit;
	vector<int>::iterator  iit;

	tree->global_alignment_action.clear();
	tree->global_alignment_fromAnc.clear();
	tree->global_alignment_indelNo.clear();

	cit = tree->global_alignment_action.begin();
	tree->global_alignment_action.insert(cit, inNumSites+1, 'i');
	iit = tree->global_alignment_fromAnc.begin();
	tree->global_alignment_fromAnc.insert(iit, inNumSites+1, -1);
	iit = tree->global_alignment_indelNo.begin();
	tree->global_alignment_indelNo.insert(iit, inNumSites+1, -1);
}

void CreateSequences(TTree *tree, int inNumSites)
{
	for (list<TNode*>::iterator itn = tree->nodeList.begin(); itn != tree->nodeList.end(); itn++) {
		try {
    		(*itn)->sequence.assign(inNumSites, (char)0);
		} catch (exception& e) {
			cerr << "Error allocating memory in CreateSequences: " << e.what() << endl;
			exit(EXIT_FAILURE);
		}
	}
}

void SetCategories(TNode *node, int inNumSites, seqGenOptions *options)
{
    int i;
    double sumRates;
	stringstream message;

	CreateRates(node, inNumSites);

	if (node->anc != NULL) {
		switch (node->nodeEnv->rateHetero) {
		case GammaRates:
			// GammaRates->GammaRates, NoRates->GammaRates
			if (node->anc->nodeEnv->rateHetero != GammaRates && node->anc->nodeEnv->rateHetero != NoRates) {
				message << "Rate heterogeneity must remain gamma rates in clade parameters: Alpha parameter can be changed.";
				options->SpoolWarnings(message.str());
				node->nodeEnv->rateHetero = node->anc->nodeEnv->rateHetero;
				if (node->anc->nodeEnv->rateHetero == DiscreteGammaRates) 
					node->nodeEnv->numCats = node->anc->nodeEnv->numCats;
				// Changed the rateHetero of the current node, reprocess.
				SetCategories(node, inNumSites, options);
				return;
			}
			if (node->nodeEnv->gammaShape == node->anc->nodeEnv->gammaShape) {
				node->gammaRates = node->anc->gammaRates;
				return;
			} // or else need to reset gamma's, as done below.
			break;
		case DiscreteGammaRates:
			// DiscreteGammaRates->DiscreteGammaRates, NoRates->DiscreteGammaRates
			if (node->anc->nodeEnv->rateHetero != DiscreteGammaRates && node->anc->nodeEnv->rateHetero != NoRates) {
				message << "Rate heterogeneity must remain discrete gamma rates in clade parameters.";
				options->SpoolWarnings(message.str());
				node->nodeEnv->rateHetero = node->anc->nodeEnv->rateHetero;
				if (node->anc->nodeEnv->rateHetero == GammaRates)
					node->nodeEnv->gammaShape = node->anc->nodeEnv->gammaShape;
				SetCategories(node, inNumSites, options);
				return;
			}
			if (node->nodeEnv->gammaShape == node->anc->nodeEnv->gammaShape &&
				node->anc->nodeEnv->rateHetero == DiscreteGammaRates) {
				node->categories = node->anc->categories;
				return;
			}
			break;
		case CodonRates:
			// CodonRates->CodonRates
			if (node->anc->nodeEnv->rateHetero != CodonRates) {
				message << "Cannot change from codon rates in lineages";
				options->SpoolWarnings(message.str());
				node->nodeEnv->rateHetero = node->anc->nodeEnv->rateHetero;
				SetCategories(node, inNumSites, options);
				return;
			}
			break;
		case NoRates:
			// NoRates->NoRates, GammaRates->NoRates, DiscreteGammaRates->NoRates
			if (node->anc->nodeEnv->rateHetero == CodonRates) {
				message << "Cannot change from Uniform Rates to Codon Rates in lineages.";
				options->SpoolWarnings(message.str());
				node->nodeEnv->rateHetero = CodonRates;
				node->nodeEnv->catRate[0] = node->anc->nodeEnv->catRate[0];
				node->nodeEnv->catRate[1] = node->anc->nodeEnv->catRate[1];
				node->nodeEnv->catRate[2] = node->anc->nodeEnv->catRate[2];
			} else if (node->anc->nodeEnv->rateHetero == NoRates) {
				return;
			} else {
				node->nodeEnv->rateHetero = node->anc->nodeEnv->rateHetero;
				SetCategories(node, inNumSites, options);
				return;
			}
			break;
		}
	}

	// If not returned from above, we're cool to change the rates.
    if (node->nodeEnv->rateHetero==CodonRates) {
		sumRates=node->nodeEnv->catRate[0]+node->nodeEnv->catRate[1]+node->nodeEnv->catRate[2];
		if (sumRates!=3.0) {
	    	node->nodeEnv->catRate[0]*=3.0/sumRates;
	    	node->nodeEnv->catRate[1]*=3.0/sumRates;
	    	node->nodeEnv->catRate[2]*=3.0/sumRates;
		}
    } else if (node->nodeEnv->rateHetero==GammaRates) {
		for (i=0; i<inNumSites; i++) {
	    	node->gammaRates.at(i)=rndgamma(node->nodeEnv->gammaShape) / node->nodeEnv->gammaShape;
		}
    } else if (node->nodeEnv->rateHetero==DiscreteGammaRates) {
		DiscreteGamma(freqRate, node->nodeEnv->catRate, node->nodeEnv->gammaShape, node->nodeEnv->gammaShape, node->nodeEnv->numCats, 0);
		for (i=0; i<inNumSites; i++) node->categories.at(i)=(int)(rndu()*node->nodeEnv->numCats);
    }

}

char SetState(double *P, string& caller)
{
    char j = 0;
    double r;
	bool done = false;
	int num_rounds = 0;
	double *original_P = P;
	
	do {
		P = original_P;
		num_rounds++;
	    r=rndu();
	    for (j=0; r>=(*P) && j<numStates; j++) P++;
		if (j < numStates) {
			done = true;
		}
		if (num_rounds > 250) {
			cerr << "Received value " << (double)(*P) << " from " << caller << ": could not find a suitable value to return in SetState." << endl;
			cerr << "Value r = " << r << " vs value *P = " << *original_P << " that was received." << endl;
			exit(EXIT_FAILURE);
		}
		
		if (!done) { cerr << "orig_P = " << *original_P << "    r = " << r << endl; }
		
	} while (!done);

    return j;
}

short IsInvariable(double proportionInvariable)
{
    double r;

    r=rndu();
    if (r < proportionInvariable) return INVARIABLE;
    else return 0;
}

void RandomSequence(char *seq, int inNumSites)
{
    int i;
    char *P;
	string calling_routine = "RandomSequence";
		
    P=seq;
    for (i=0; i<inNumSites; i++) {
		*P=SetState(addFreq, calling_routine);
		P++;
    }
}

string insertFillSequence(int insertSize)
{
    int i,j;
    vector<double> bayes_probabilities_for_next (numStates, 0);
    double r;
	char *P;
	string insert_sequence (insertSize+1, '\0');
	string calling_routine;
	
	if (!isNucModel) {
		calling_routine.assign("insertFillSequence, isNucModel");
	    insert_sequence.at(0) = SetState(addFreq, calling_routine);
	    for(i=1; i<insertSize+1; i++) {
			bayes_probabilities_for_next.at(0)=(freq[0]*NeighborRates[(int)insert_sequence.at(i-1)*numStates])/(freq[insert_sequence.at(i-1)]);
	        for(j=1;j<numStates;j++) {
		    	bayes_probabilities_for_next.at(j)=bayes_probabilities_for_next.at(j-1);
		    	bayes_probabilities_for_next.at(j)+=(freq[j]*NeighborRates[(int)insert_sequence.at(i-1)*numStates+j])/(freq[insert_sequence.at(i-1)]);
			}

			r=rndu();
			r*=bayes_probabilities_for_next.at(numStates-1);
			j=0;
			while(r>bayes_probabilities_for_next.at(j)) j++; 
			if (j >= numStates) {
				cerr << "Illegal character found in iFS: " << j << endl;
				exit(EXIT_FAILURE);
			}
			insert_sequence.at(i) = j;
	    }
	} else {
		for (i = 0; i < insertSize+1; i++) {
			calling_routine.assign("insertFillSequence, isNucModel");
			insert_sequence.at(i) = SetState(addFreq, calling_routine);
			if (insert_sequence.at(i) >= numStates) {
				cerr << "illegal character found for insertSize in iFS: " << insert_sequence.at(i) << endl;
				exit(EXIT_FAILURE);
			}
		}
	}

	return insert_sequence;
}

void MutateSequence(inTree *iTree, TNode *des, string& seq, double inlen, double motif_correction, list<eventTrack*> *events)
{
    int i=0, j, cat, pos;
    char result;
    double len, in_value;
	bool do_sub = false;
	string calling_routine;

    switch (des->nodeEnv->rateHetero) {
	case GammaRates:
	    if (des->nodeEnv->invariableSites) {
			for (int i = 0; i < seq.size(); i++) {
				pos = (double)rndu() * (seq.size()-1);
				if ( !des->myMotif.at(pos)->activeProps->subst->siteInvariable() ) {
					len = inlen + motif_correction;
					in_value = des->gammaRates.at(pos)*len;
					SetVector(cvector, seq.at(pos), in_value);
					calling_routine.assign("MutateSequence, GammaRates, invariableSites");
					result=SetState(cvector, calling_routine);
//				cerr << "MS: result: " << stateCharacters[result] << endl; 
					// If not part of the motif, then change regardless of result value.
					if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) ) 
						seq.at(pos) = result;
					// Otherwise, if it is part of the motif, and it is another accepting value, accept.
					else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
						seq.at(pos) = result;
//					else cerr << pos << ": Rejected " << stateCharacters[result] << " at position accepting " << des->myMotif.at(pos)->activeProps->subst->report_bitset() << endl;
		    	}
			}
	    } else {
			for (int i = 0; i < seq.size(); i++) {
				pos = (double)rndu() * (seq.size()-1);
				len = inlen + motif_correction;
				in_value = des->gammaRates.at(pos)*len;
		    	SetVector(cvector, seq.at(pos), in_value);
				calling_routine.assign("MutateSequence, GammaRates");
				result=SetState(cvector, calling_routine);
				//cerr << "MS: result: " << stateCharacters[result] << endl; 
				if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) ) 
					seq.at(pos) = result;
				else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
					seq.at(pos) = result;
//				else cerr << pos << ": Rejected " << stateCharacters[result] << " at position accepting " << des->myMotif.at(pos)->activeProps->subst->report_bitset() << endl;
			}
	    }
	    break;

	case DiscreteGammaRates:
	    for (i=0; i < des->nodeEnv->numCats; i++)
	    	SetMatrix(matrix[i], des->nodeEnv->catRate[i]*(inlen+motif_correction));
	    if (des->nodeEnv->invariableSites) {
			for (int i = 0; i < seq.size(); i++) {
				pos = (double)rndu() * (seq.size()-1);

				if ( !des->myMotif.at(pos)->activeProps->subst->siteInvariable() ) {
					calling_routine.assign("MutateSequence, DiscreteGammaRates, invariableSites");
					result=SetState(matrix[des->categories.at(pos)]+(seq.at(pos)*numStates), calling_routine);
					if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) ) 
						seq.at(pos) = result;
					else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
						seq.at(pos) = result;
				}
			}
	    } else {
			for (int i = 0; i < seq.size(); i++) {
				pos = (double)rndu() * (seq.size()-1);
				calling_routine.assign("MutateSequence, DiscreteGammaRates");
				result=SetState(matrix[des->categories.at(pos)]+(seq.at(pos)*numStates),calling_routine);
				if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) ) 
					seq.at(pos) = result;
				else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
					seq.at(pos) = result;
//				else cerr << pos << ": Rejected " << stateCharacters[result] << " at position accepting " << des->myMotif.at(pos)->activeProps->subst->report_bitset() << endl;
			}
	    }
	    break;

	case CodonRates:
	    for (i=0; i < 3/*des->nodeEnv->numCats*/; i++)
	    	SetMatrix(matrix[i], des->nodeEnv->catRate[i]*inlen);
		for (int i = 0; i < seq.size(); i++) {
			pos = (double)rndu() * (seq.size()-1);
			cat=(pos+iTree->codon_offset)%3;
			calling_routine.assign("MutateSequence, CodonRates");
			result=SetState(matrix[cat]+(seq.at(pos) * numStates), calling_routine);
			if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) ) 
				do_sub=true;
			else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
				do_sub = true;

			if (do_sub) {
				// Check if stop codon is result of substitution.
				int codon[3];
				if (pos - (3-iTree->codon_offset) < 0 || pos + (3-iTree->codon_offset) >= seq.size()) {
					; // Punt, since I do not keep track of other partitions (what about subsequences??)
				} else if ( (pos+iTree->codon_offset) % 3 == 0 ) {
					codon[0] = result;
					codon[1] = seq.at(pos+1);
					codon[2] = seq.at(pos+2);
				} else if ( (pos+iTree->codon_offset) % 3 == 1 ) {
					codon[0] = seq.at(pos-1);
					codon[1] = result;
					codon[2] = seq.at(pos+1);
				} else if ( (pos+iTree->codon_offset) % 3 == 2 ) {
					codon[0] = seq.at(pos-2);
					codon[1] = seq.at(pos-1);
					codon[2] = result;
				} else {
					cerr << "Huh?? how can there be more than 3 codon positions?" << endl;
					exit(EXIT_FAILURE);
				}
				if ( !Stop_Codon(codon) ) seq.at(pos) = result;
//				else cerr << "rejected stop codon: " << codon[0] << codon[1] << codon[2] << endl;
			}
			// else cerr << pos << ": Rejected " << stateCharacters[result] << " at position accepting " << des->myMotif.at(pos)->activeProps->subst->report_bitset() << endl;
	    }
	    break;

	case NoRates:
	    SetMatrix(matrix[0], inlen+motif_correction);
	    if (des->nodeEnv->invariableSites) {
			for (int i = 0; i < seq.size(); i++) {
				pos = (double)rndu() * (seq.size()-1);
				if ( !des->myMotif.at(pos)->activeProps->subst->siteInvariable() ) {
					calling_routine.assign("MutateSequence, NoRates, invariableSites");
					result=SetState(matrix[0]+(seq.at(pos)*numStates), calling_routine);
					if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) )
						seq.at(pos) = result;
					else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
						seq.at(pos) = result;
//					else cerr << pos << ": Rejected " << stateCharacters[result] << " at position accepting " << des->myMotif.at(pos)->activeProps->subst->report_bitset() << endl;
				}
			}
	    } else {
			for (int i = 0; i < seq.size(); i++) {
				pos = (double)rndu() * (seq.size()-1);
				calling_routine.assign("MutateSequence, NoRates");
		    	result=SetState(matrix[0]+(seq.at(pos)*numStates), calling_routine);
				if ( !(des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(seq.at(pos))) ) 
					seq.at(pos) = result;
				else if (des->myMotif.at(pos)->activeProps->subst->substitution_bitstring.test(result) )
					seq.at(pos) = result;
//				else cerr << pos << ": Rejected " << stateCharacters[result] << " at position accepting " << des->myMotif.at(pos)->activeProps->subst->report_bitset() << endl;
			}
	    }
	    break;
    }
}

bool Stop_Codon(int *codon) 
{
	size_t T = stateCharacters.find("T"), A = stateCharacters.find("A"), G = stateCharacters.find("G");

	if (codon[0] == T && codon[1] == A && codon[2] == G) return true;
	if (codon[0] == T && codon[1] == G && codon[2] == A) return true;
	if (codon[0] == T && codon[1] == A && codon[2] == A) return true;

	return false;
}

void InDel(TTree *tree, TNode  *des, double len, list<eventTrack*> *events, int CnB_divisor) 
{
    int action = INSERT;
    int indel_size;
    int i;
	int E_in_ = 0;
	int E_del_ = 0;
	int total_trials;

	total_trials = calcTrials(des,&E_in_,&E_del_);		// Calculate rounds by examining indel objects.

	bool success;
	for(i=0; i<total_trials; i++) {
		if (des->nodeEnv->rateHetero == CodonRates) {
			if ((double)rndu() < (double)E_in_ / ((double)E_in_+E_del_))
				action = Find_Action(len,des->nodeEnv->P_ins_ / 3.,des->nodeEnv->P_del_ / 3., CnB_divisor, INSERT); 
			else
				action = Find_Action(len,des->nodeEnv->P_ins_ / 3.,des->nodeEnv->P_del_ / 3., CnB_divisor, DELETE); 
		} else {
			if ((double)rndu() < (double)E_in_ / ((double)E_in_+E_del_))
				action = Find_Action(len,des->nodeEnv->P_ins_,des->nodeEnv->P_del_, CnB_divisor, INSERT); 
			else
				action = Find_Action(len,des->nodeEnv->P_ins_,des->nodeEnv->P_del_, CnB_divisor, DELETE); 
		}
		
		if(action == INSERT || action == DELETE) {
			eventTrack *new_event = new eventTrack(INSERT | DELETE);
			success = false;
	    	if(action == INSERT) {
	        	indel_size=Find_Indel_Size(des->nodeEnv->insert_lengthDistribution,des->nodeEnv->maxIndel);	
				if ( (success=Insert(tree,des,indel_size)) ) {
					num_insert++;
					len_insert += indel_size;
				}
	    	} else {
	        	indel_size=Find_Indel_Size(des->nodeEnv->delete_lengthDistribution,des->nodeEnv->maxIndel);
				if ( (des->sequence.size()-indel_size) > 0)
		    		if ( (success=Delete(tree,des,indel_size)) ) {
		    			num_delete++;
						len_delete += indel_size;
					}
	    	}
			if (success) {
		    	new_event->NewEvent(tree,indelNo,action,des->atEpochTime,des->bipartition,indel_size);
				(*events).push_back(new_event);
				indelNo++;
			}
		}
    }
	if (PRINT_ID) {
		cout << des->atEpochTime << " "; 
		cout << num_insert << " " << len_insert << " " << num_delete << " " << len_delete << endl;
	}
	
	for (vector<motifSite*>::iterator it = des->myMotif.begin(); it != des->myMotif.end(); it++) {
		(*it)->Print_Nulls();
	}
}

int calcTrials(TNode *des, int *E_in_, int *E_del_)
{
	// Do we need to make an adjustment to the trial counting for regexes? (esp. cxxc type)
	if(des->nodeEnv->invariableSites) {
		for (vector<motifSite*>::iterator it = des->myMotif.begin(); it != des->myMotif.end(); it++) {

			// Insertion calculations:
			if ( (*it)->activeProps->indel->L_ins_->insertionAllowed() ) {
				(*E_in_)++;
			}

			// Deletion  calculations:		
			if ( (*it)->activeProps->indel->del->deletionAllowed() ) {
				(*E_del_)++;
			}

			// Last site right needs to be checked.
			if ( it+1 == des->myMotif.end() ) {
				if ( (*it)->activeProps->indel->R_ins_->insertionAllowed() ) (*E_in_)++;
			}
		}
	} else {
		*E_in_ = des->sequence.size()+1;
		*E_del_ = des->sequence.size();
	}

	// C&B are half n half.
	if (des->nodeEnv->P_ins_ == 0 && des->nodeEnv->P_del_ == 0) 
		*E_in_ = *E_del_ = (*E_in_+*E_del_) / 2;

	return *E_in_ + *E_del_;
}

int  Find_Action(double len, double P_insert_, double P_delete_, int CnB_divisor, int action_test) 
{
    double P_indel_; 
  	int action;
  
    if( !P_insert_ && !P_delete_) {
        P_indel_ = (0.0224 - 0.0219 * exp(-0.01168*(len*100))) / (double)CnB_divisor;
		if((double)rndu() < P_indel_) {
            if((double)rndu() < 0.5) return INSERT;
            else return DELETE;
        } else {
        	return NO_ACTION;
		}
    } else {
		if (action_test == INSERT) {
			if ((double)rndu() < P_insert_*len) return INSERT;
		} else  {
			double chkval = (double)rndu();
			if (chkval < P_delete_*len) {
				return DELETE;
			}
		}
    } 
    return NO_ACTION;
} 

void Find_Motif_Positions(TTree *tree, TNode *des, motifSite *this_site, bool back, int start_aa, varSite **in_template_varSite, varSite **in_motif_varSite)
{
	vector<motifSite*>::iterator mSit;

	// Inserted. Now need to make each one the activeProp, and patch each end to the correct varSites
	// in the existing sequence.
	mSit = des->myMotif.begin();
	// Interesting problem arises if we simply use start_aa: Since we are not looking between sites
	// when calling motifSites, we will actually set the incorrect values for the varSites; we
	// will set the varSites to the next position, which is a major problem when flanking a region
	// that does not accept insertions. Thus, we will get the } else { statement instead of the
	// if or if-else stmt.
	bool st_L = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->insertion();
	bool m_L = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->insertion();
	bool st_R = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->insertion();
	bool m_R = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]->insertion();
	bool os_st = false, os_m = false;
	if (this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite)
		os_st = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite->insertion();
	if (this_site->activeProps->indel->R_ins_->on_site_motif_varSite)
		os_m = this_site->activeProps->indel->R_ins_->on_site_motif_varSite->insertion();

	// For the randomization of finding which cases I try. Each time I try a case, I set that bit
	// to a '1' to avoid infinite loop (if it is even truly a possibility).
	bitset<6> tried_case (string("000000"));

//cerr << "pos " << start_aa << ": ";
	
	double random = (double)rndu(), random2;
	if ( back ) {
		*in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
		*in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
	} else {
		bool found_site = false;
		int num_rounds = 0;

		// Randomize the position of the varSite where insertion occurs.
		while(!found_site) {
			found_site = true;
			random2 = (double)rndu();
			size_t rand = (size_t)(random2 * 6.0);

			if(num_rounds++ > 250) {
				cerr << "case left undealt with in Insert routine:" << endl;
				cerr << "st_L " << st_L << " st_R " << st_R << " m_L " << m_L << " m_R " << m_R << " os_st " << os_st << " os_m " << os_m << endl;
				exit(EXIT_FAILURE);
			}

			// If already tested, no need to re-invent the wheel (i.e., it still won't work).
			if (!tried_case.test(rand)) {
				tried_case.set(rand, true);
				switch (rand) {

				case 0:
//					cerr << "case 0" << endl;
					if (st_L) {
						*in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
						if (m_L && os_m)
							if (random < 0.5) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
							else *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (os_m) *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (m_L) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
						else found_site = false;
					} else found_site = false;
					break;

				case 1:
//					cerr << "case 1" << endl;
					if (st_R) {
						*in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
						if (m_R && os_m)
							if (random < 0.5) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
							else *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (os_m) *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (m_R) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
						else found_site = false;
					} else found_site = false;
					break;

				case 2:
//					cerr << "case 2" << endl;
					if (m_R) {
						*in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
						if (st_R && os_st)
							if (random < 0.5) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
							else *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else if (st_R) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
						else if (os_st) *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else found_site = false;
					} else found_site = false;
					break;

				case 3:
//					cerr << "case 3" << endl;
					if (m_L) {
						*in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
						if (st_L && os_st)
							if (random < 0.5) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
							else *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else if (st_L) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
						else if (os_st) *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else found_site = false;
					} else found_site = false;
					break;

				case 4:
//					cerr << "case 4" << endl;
					if (os_st) {
						*in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						if (m_L && m_R && os_m) 
							if (random < 0.333) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
							else if (random < 0.666) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
							else *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (m_L && m_R) 
							if (random < 0.5) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
							else *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
						else if (m_L && os_m) 
							if (random < 0.5) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
							else *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (m_R && os_m) 
							if (random < 0.5) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
							else *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else if (m_L) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT];
						else if (m_R) *in_motif_varSite = this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT];
						else if (os_m) *in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						else found_site = false;
					} else found_site = false;
					break;

				case 5:
//					cerr << "case 5" << endl;
					if (os_m) {
						*in_motif_varSite = this_site->activeProps->indel->R_ins_->on_site_motif_varSite;
						if (st_L && st_R && os_st) 
							if (random < 0.333) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
							else if (random < 0.666) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
							else *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else if (st_L && st_R) 
							if (random < 0.5) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
							else *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
						else if (st_L && os_st) 
							if (random < 0.5) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
							else *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else if (st_R && os_st) 
							if (random < 0.5) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
							else *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else if (st_L) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
						else if (st_R) *in_template_varSite = this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
						else if (os_st) *in_template_varSite = this_site->activeProps->indel->R_ins_->on_site_sequence_template_varSite;
						else found_site = false;
					} else found_site = false;
					break;

				default:
					// should be impossible.
					des->Print_Active_Sites(des->output_sequence(des->sequence.begin(), des->sequence.end()));
					cerr << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->insertion() << " " << "(" << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->min << "," << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->max << ") " << this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->insertion() << "  " << "(" << this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->min << "," << this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->max << ")  " << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->insertion() << " " << "(" << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->min << "," << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->max << ") " << this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]->insertion() << " " << "(" << this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]->min << "," << this_site->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]->max << ")" << endl << "members.size() = " << this_site->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->members.size() << " " << this_site->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->members.size() << " " << "Wrong insertion site????" << endl;
					exit(EXIT_FAILURE);
					break;
				}
			} else found_site = false;
		}
	}

}

int  Insert(TTree *tree, TNode *des, int indel_size) 
{
    int start_aa;
    int final_aa;
    int i,j;
    int indel_aa_in_globals=0;
    int position_in_globals,num_indels_in_globals;
	vector<char>::iterator   cit;
	vector<int>::iterator    iit;
	vector<short>::iterator  sit;
	vector<double>::iterator dit;
	vector<motifSite*>::iterator mSit;

	position_in_globals = 0;
	num_indels_in_globals = 0;

    final_aa=des->sequence.size();
    start_aa=Find_Start_Location(des,&indel_size,des->sequence.size(),INSERT);

	if (start_aa == FSL_FAIL) return 0;

	// For insertion, working on half sites, which in turn means we have to go to the next site
	// to represent an insertion correctly. 
	start_aa++;

    if(start_aa > final_aa) {
		fprintf(stderr,"start_aa=%d\t+\tindel_size=%d\t>\tfinal_aa=%d\n",start_aa,indel_size,final_aa);
		exit(EXIT_FAILURE);
    }

    while(indel_aa_in_globals <= start_aa) { 
        if(position_in_globals > tree->global_alignment_action.size()) { 
            fprintf(stderr,"INSERT: Over g_a_end.  indel_aa_in_globals = %d,\tstart_aa = %d\n", indel_aa_in_globals,start_aa);
			cerr << "tree->global_alignment_action.size() " << tree->global_alignment_action.size() << " position_in_globals " << position_in_globals << endl;
            exit(EXIT_FAILURE); 
        } 
        if(isSpot(tree,&position_in_globals,des)) { indel_aa_in_globals++; } 
        position_in_globals += 1; 
    } 

    num_indels_in_globals = position_in_globals; 
    while(indel_aa_in_globals <= start_aa+indel_size) {
		indel_aa_in_globals++;
        num_indels_in_globals += 1; 
    } 

	cit = tree->global_alignment_action.begin();
	tree->global_alignment_action.insert(cit+position_in_globals, indel_size, 'i');
	iit = tree->global_alignment_fromAnc.begin();
	tree->global_alignment_fromAnc.insert(iit+position_in_globals, indel_size, des->mytipNo);
	iit = tree->global_alignment_indelNo.begin();
	tree->global_alignment_indelNo.insert(iit+position_in_globals, indel_size, indelNo);

	// Insert into existing sequence.
	string insert_sequence = insertFillSequence(indel_size);

	if (start_aa >= des->sequence.size()) {
		cerr << "This could be a problem causing case." << endl;
		des->sequence.append(insert_sequence,0,indel_size);
		if (des->nodeEnv->invariableSites)
			des->invariable.insert(des->invariable.end(), indel_size, 0);
	} else {
		if (des->nodeEnv->invariableSites) {
			sit = des->invariable.begin();
			des->invariable.insert(sit+start_aa, indel_size, 0);
		}
		des->sequence.insert(start_aa,insert_sequence,0,indel_size);
	}
	
	/////// Make this a function??? This is quite large....

	mSit = des->myMotif.begin();
	varSite *in_template_varSite, *in_motif_varSite;
	motifSite *this_site = (*(mSit+start_aa-1));
	bool back = ( ( this_site == des->myMotif.back()) ? true : false );
	Find_Motif_Positions(tree, des, this_site, back, start_aa, &in_template_varSite, &in_motif_varSite);
	
	if (in_template_varSite == NULL || in_motif_varSite == NULL) {
		cerr << "Unsuccessful in locating correct insertion position." << endl;
		exit(EXIT_FAILURE);
	}

	vector<motifSite*> newSites;
	for (i = 0; i < indel_size; i++) 
		newSites.push_back(new motifSite(this_site, in_template_varSite, in_motif_varSite));

	// Link newSites & set activeProps.
	// * Middle sites:
	delete newSites.front()->activeProps->indel->L_ins_;
	if (newSites.size() > 1) {	
		for (vector<motifSite*>::iterator it = newSites.begin()+1; it != newSites.end(); it++) {
			delete (*it)->activeProps->indel->L_ins_;
			(*it)->activeProps->indel->L_ins_ = (*(it-1))->activeProps->indel->R_ins_;
		}
	}
	
	// * Insert the newSites into the motifSites.
	if ( back ) 
		for (vector<motifSite*>::iterator it = newSites.begin(); it != newSites.end(); it++)
			des->myMotif.push_back(*it);
	else des->myMotif.insert(mSit+start_aa, newSites.begin(), newSites.end());

	vector<motifSite*>::iterator first_inserted_site = des->myMotif.begin()+start_aa;

	// * Now, link entry points: 
	// * * First site:
	(*first_inserted_site)->activeProps->indel->L_ins_ = (*(first_inserted_site-1))->activeProps->indel->R_ins_;
	(*first_inserted_site)->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT] = in_template_varSite;
	(*first_inserted_site)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT] = in_motif_varSite;

	// * * Last site:
	if ( !back ) (*(first_inserted_site+newSites.size()))->activeProps->indel->L_ins_ = (*(first_inserted_site+newSites.size()-1))->activeProps->indel->R_ins_;
	(*(first_inserted_site+newSites.size()-1))->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT] = in_template_varSite;
	(*(first_inserted_site+newSites.size()-1))->activeProps->indel->R_ins_->my_motif_varSite[RIGHT] = in_motif_varSite;

	// Done linking. Should not be differentiable from other sites.
	des->postProcess();

	// Place new sites into varSite's members array.
	size_t check = 0;
	for (vector<motifSite*>::iterator it = first_inserted_site; it != first_inserted_site+indel_size; it++) {
		(*it)->activeProps->indel->del->my_sequence_template_varSite->add2Member((*it)->activeProps->indel->del);
		(*it)->activeProps->indel->del->my_motif_varSite->add2Member((*it)->activeProps->indel->del);
	}

	vector<double> d_insert_vector;
	vector<short>  s_insert_vector;
	switch (des->nodeEnv->rateHetero) {
	case GammaRates:
		dit = des->gammaRates.begin();
		des->gammaRates.insert(dit+start_aa, indel_size, 0);
		for (i=start_aa; i<start_aa+indel_size; i++)
			des->gammaRates.at(i) = rndgamma(des->nodeEnv->gammaShape) / des->nodeEnv->gammaShape;
		break;
	case DiscreteGammaRates:
		s_insert_vector.clear();
		for (i=start_aa; i<start_aa+indel_size; i++)
			s_insert_vector.push_back((int)(rndu()*des->nodeEnv->numCats));
		sit = des->categories.begin();
		des->categories.insert(sit+start_aa, s_insert_vector.begin(), s_insert_vector.end());
		break;
	case CodonRates:
		// Taken care of in mutatesequences
		break;
	case NoRates:
		break;
	}
	
	return 1;
}

int Delete(TTree *tree,TNode *des, int indel_size) 
{
    int start_aa;
    int final_aa;
    int i,j=0;
    int indel_aa_in_globals = 0; 
    int position_in_globals; 
    int num_indels_in_globals; 
    int temp_in_globals; 
    int t_i_g_prev = 0;
    int num_rounds = 0; 
	vector<int>::iterator    iit;
	vector<char>::iterator   cit;
	vector<short>::iterator  sit;
	vector<double>::iterator dit;

    final_aa=des->sequence.size();
    start_aa=Find_Start_Location(des,&indel_size,des->sequence.size(),DELETE);

	if (start_aa == FSL_FAIL) return 0;

 	// Find corresponding position in globals...
    position_in_globals = 0;
    num_indels_in_globals = 0;
 
    while(indel_aa_in_globals <= start_aa) { 
        if(position_in_globals > tree->global_alignment_action.size()) {
            printf("DELETE: Over g_a_end.  indel_aa_in_globals = %d,\tstart_aa = %d\tfinal_aa = %d\tdelete_size = %d\n", 
		            indel_aa_in_globals,start_aa,final_aa,indel_size);
		    cerr << "position_in_globals = " << position_in_globals << " tree->global_alignment_action.size() = " << tree->global_alignment_action.size() << endl;
            exit(EXIT_FAILURE);
        }
        if(isSpot(tree,&position_in_globals,des)) { indel_aa_in_globals++; }
        position_in_globals += 1;
    }

    num_indels_in_globals = position_in_globals;
    while(indel_aa_in_globals <= start_aa+indel_size) {
        if(isSpot(tree,&num_indels_in_globals,des)) { indel_aa_in_globals++; num_rounds++; }
        num_indels_in_globals += 1; 
    } 

    temp_in_globals = position_in_globals;
    t_i_g_prev = temp_in_globals;
    for(i = position_in_globals; i < num_indels_in_globals+num_rounds; i++) { 
        temp_in_globals = i;
        if(isSpot(tree,&temp_in_globals,des)) { 
			cit = tree->global_alignment_action.begin();
			tree->global_alignment_action.insert(cit+i, 1, 'd');
			iit = tree->global_alignment_fromAnc.begin();
			tree->global_alignment_fromAnc.insert(iit+i, 1, des->mytipNo);
			iit = tree->global_alignment_indelNo.begin();
			tree->global_alignment_indelNo.insert(iit+i, 1, indelNo);
            temp_in_globals += 1;
        }    
        i = temp_in_globals;
    }

    if(delete_debug) Print_Globals(tree); 
	des->sequence.erase(start_aa, indel_size);

	if (des->nodeEnv->invariableSites) {
		sit = des->invariable.begin();
		des->invariable.erase(sit+start_aa, sit+(start_aa+indel_size));
	}

	vector<motifSite*>::iterator tit = des->myMotif.begin()+start_aa;
	motifSite *next;
	for (size_t num_dels = 0; num_dels != indel_size; num_dels++) {
		if ( (*tit) == des->myMotif.back() ) next = NULL; 
		else next = (*(tit+1));
		(*tit)->deleteMerge( (*(tit-1)), next );
		delete (*tit);
		des->myMotif.erase(tit);
	}

	// Will need to change this to lineage-specific...
	switch (des->nodeEnv->rateHetero) {
	case GammaRates:
		dit = des->gammaRates.begin();
		des->gammaRates.erase(dit+start_aa, dit+(start_aa+indel_size));
		break;
	case DiscreteGammaRates:
		sit = des->categories.begin();
		des->categories.erase(sit+start_aa, sit+(start_aa+indel_size));
		break;
	case CodonRates:
		// Taken care of in mutatesequences
		break;
	case NoRates:
		break;
	}
	
	return 1;
} 

int Find_Start_Location(TNode *des, /*TTree *tree,*/ int *indel_size, int final_aa, int action) 
{
    int position;
    int i,j; 
    int ok; 
    int numAcceptablePositions = 0;
    vector<bool> acceptable_positions (final_aa+2, false);
	vector<bool>::iterator bit = acceptable_positions.begin();

    if(action == INSERT) {
//		if (des->nodeEnv->rateHetero == CodonRates && tree->codons_only) {

//		} else {
			for (vector<motifSite*>::iterator it = des->myMotif.begin(); it != des->myMotif.end(); it++, bit++) {
				// Insertion calculations:
				if ( (*it)->activeProps->indel->R_ins_->insertionAllowed(*indel_size) ) {
					(*bit) = true;
					numAcceptablePositions++;
				}
			}
//		}
 	} else { 
		// Two arrays, if a deletion of indel_size is acceptable for each structure, then this pos
		// is an acceptable pos. If not, no dice.
 		vector<bool> st_accept (final_aa+2, false);
 		vector<bool> m_accept (final_aa+2, false);
		varSite *var_site, *current_varSite;
		size_t members, min, current_indel_size, num_dels_in_current_varSite, max;
		bool success;
		vector<bool>::iterator pos;
		pos = st_accept.begin();
		for (vector<motifSite*>::iterator it = des->myMotif.begin(); it != des->myMotif.end(); it++, pos++) {
			// Template array fillage.
			current_varSite = NULL;
			current_indel_size = 0;
			success = true;
			for (vector<motifSite*>::iterator it2 = it; it2 != des->myMotif.end() && current_indel_size != *indel_size; it2++, current_indel_size++) {
				// If we are in a new varSite, need to reflect that in all data members.
				if ( (*it2)->activeProps->indel->del->my_sequence_template_varSite != current_varSite) {	
					current_varSite = (*it2)->activeProps->indel->del->my_sequence_template_varSite;
					members = current_varSite->members.size();
					min = current_varSite->min;
					max = current_varSite->max;
					num_dels_in_current_varSite = 0;
				}
				// Too many dels in this site, less than minimum acceptable size.
				num_dels_in_current_varSite++;
				if (members - num_dels_in_current_varSite < min || max == min ) success = false;
			}
			// Near the end of the sequence.
			if (current_indel_size < *indel_size) success = false;
			(*pos) = success;
		}

		pos = m_accept.begin();
		for (vector<motifSite*>::iterator it = des->myMotif.begin(); it != des->myMotif.end(); it++, pos++) {
			// Motif array fillage.
			current_varSite = NULL;
			current_indel_size = 0;
			success = true;
			for (vector<motifSite*>::iterator it2 = it; it2 != des->myMotif.end() && current_indel_size != *indel_size; it2++, current_indel_size++) {
				// If we are in a new varSite, need to reflect that in all data members.
				if ( (*it2)->activeProps->indel->del->my_motif_varSite != current_varSite) {	
					current_varSite = (*it2)->activeProps->indel->del->my_motif_varSite;
					members = current_varSite->members.size();
					min = current_varSite->min;
					max = current_varSite->max;
					num_dels_in_current_varSite = 0;
				}
				// Too many dels in this site, less than minimum acceptable size.
				num_dels_in_current_varSite++;
				if (members - num_dels_in_current_varSite < min || max == min ) success = false;
			}
			// Near the end of the sequence.
			if (current_indel_size < *indel_size) success = false;
			(*pos) = success;
		}

		vector<bool>::iterator it2 = m_accept.begin();
		bit = acceptable_positions.begin();
		for (vector<bool>::iterator it = st_accept.begin(); it != st_accept.end(); it++, it2++, bit++) {
			(*bit) = (*it) && (*it2);
			if (*bit) numAcceptablePositions++;
		}
    } 

	//cerr << ((action == INSERTION) ? "Insert" : "Delete") << " size " << *indel_size << endl;
	//cerr << "Acceptable Positions: ";
	//for (vector<bool>::iterator it = acceptable_positions.begin(); it != acceptable_positions.end(); it++)
	//	cerr << (*it);
	//cerr << endl;

    if(numAcceptablePositions < 0) {
		fprintf(stderr,"evolve.c, Find_Start_Locations:\nNegative number of acceptable positions.\n");
		fprintf(stderr,"numAcceptablePositions = %d\n",numAcceptablePositions);
		exit(EXIT_FAILURE);
    }

    ok=0; 
    i=0; 
    while(!ok && i<final_aa) { 
        if(acceptable_positions.at(i)) { ok=1; } 
        i++; 
    } 

    if(i >= final_aa) {
        fprintf(stderr,"There is no acceptable position to make %s of size %d.\n",
	        ((action)?"deletion":"insertion"),*indel_size);
	    des->report_varSites();
	    cerr << endl << "Sequence size = " << des->sequence.size() << endl;
		num_template_violations++;
		return FSL_FAIL;
    } else {
	    i=(int)(rndu()*(double)numAcceptablePositions+1); 
	    position=0; j=0;
	    while(!acceptable_positions.at(j) && j < acceptable_positions.size()) j++;
	    while(position < i && j < final_aa) {
	        if(acceptable_positions.at(j)) position++;
	        if(position < i) j++;
	    } 
	}
	// ::::::TESTING:::::: Why are post-motif positions giving errors?
//	cerr << ((action==INSERT)? "insert:" : "delete:") << " ";
//	cerr << "Checking site " << j-1 << " in a_p, size: " << acceptable_positions.size() << endl;
	if (j != 0) {
		if (!acceptable_positions.at(j-1)) {
			cerr << ((action==INSERT)?"insert":"delete") << " (size " << *indel_size << ") directly after non-accepting site: sequence position " << j << endl;
		}
	}
	//cerr << "returning location " << j << " for sequence size " << des->sequence.size() << endl;

    return j; 
} 
                     
int Find_Indel_Size(vector<double>& dist, int max_indel) 
{
    int size=0;   
    int i=1;  
    double sum=0;
    double rand; 
     
    rand=rndu(); 
    if(rand==0) rand=rand+0.00001; 
     
    while(i <= max_indel && size == 0) { 
        sum+=dist.at(i);
        if(sum > rand) size=i; 
        i++; 
    } 
    return size;
}        
     
int  isAnc(TNode *thisNode, int anc) 
{
    if(thisNode -> mytipNo == anc) {
        return 1; 
    } else if(thisNode -> mytipNo == -1) {
        return 0; 
    } else { 
        return isAnc(thisNode->anc, anc);
    } 
} 
             
int  isSpot(TTree *tree, int *pos, TNode *curr_taxa) 
{
    int new_pos;

    while(tree->global_alignment_action.at(*pos) != 'i' && *pos < tree->global_alignment_action.size())
    	*pos+=1;
  
    if(!isAnc(curr_taxa,tree->global_alignment_fromAnc.at(*pos))) { return 0; }
    else { 
        new_pos = *pos-1;
        while(new_pos > 0 && tree->global_alignment_action.at(new_pos) != 'i') { new_pos--; }
         
        new_pos++;
             
        if(new_pos == *pos) { return 1; } 
        else {
            while(tree->global_alignment_action.at(new_pos) == 'd') {
                if(isAnc(curr_taxa,tree->global_alignment_fromAnc.at(new_pos))) { return 0; } 
                new_pos++;
            } 
            return 1; 
        } 
    } 
} 

void initializeCladeParameters(TTree *tree, TNode *node) 
{
	inClade *env;
	if (node->anc == NULL) env = tree->treeEnv.front();
	else env = node->anc->nodeEnv;

	if (!(node->nodeEnv->set_in_clade[__indelFlag__])) {
		node->nodeEnv->indelFlag = env->indelFlag;
		if (env->indelFlag) {
			if (!(node->nodeEnv->set_in_clade[__P_ins___]))
				node->nodeEnv->P_ins_ = env->P_ins_;

			if (!(node->nodeEnv->set_in_clade[__P_del___]))
				node->nodeEnv->P_del_ = env->P_del_;

			if (!(node->nodeEnv->set_in_clade[__maxIndel__]))
				node->nodeEnv->maxIndel = env->maxIndel;

			if (!(node->nodeEnv->set_in_clade[__insert_lengthDistribution__]))
				node->nodeEnv->insert_lengthDistribution = env->insert_lengthDistribution;

			if (!(node->nodeEnv->set_in_clade[__delete_lengthDistribution__])) 
					node->nodeEnv->delete_lengthDistribution = env->delete_lengthDistribution;
		}
	}

	if (!(node->nodeEnv->set_in_clade[__values2Export2Freq__])) {
		// Model frequencies should be used if the model is set.
		if (node->nodeEnv->set_in_clade[__model__] && !isNucModel) {
			node->nodeEnv->values2Export2Freq.assign(numStates,0);
			SetModelFreqs(node->nodeEnv->model, node->nodeEnv);
		}
		else node->nodeEnv->values2Export2Freq = env->values2Export2Freq;
	}

	if (!(node->nodeEnv->set_in_clade[__model__])) 
		node->nodeEnv->model = env->model;

	if (!(node->nodeEnv->set_in_clade[__invariableSites__]))
		node->nodeEnv->invariableSites = env->invariableSites;

	if (!(node->nodeEnv->set_in_clade[__proportion_invariable__]))
		node->nodeEnv->proportion_invariable = env->proportion_invariable;

	if (!(node->nodeEnv->set_in_clade[__rateHetero__])) {
		node->nodeEnv->rateHetero=env->rateHetero;
		node->nodeEnv->gammaShape = env->gammaShape;
		node->nodeEnv->numCats = env->numCats;
		for (int i = 0; i < MAX_RATE_CATS; i++) 
			node->nodeEnv->catRate[i]=env->catRate[i];
	}
	
	if (!(node->nodeEnv->set_in_clade[__constraintChange__])) 
		node->nodeEnv->constraintChange = env->constraintChange;
}

void initializeBranchRun(inTree *iTree, TNode *node, double *en_invar_scale, int inNumSites, seqGenOptions *options) 
{
	SetModel(node->nodeEnv->model, node->nodeEnv);

	if (node->nodeEnv->constraintChange == PSEUDOGENE) {
		// Remove all constraints from lineage. It's free-flowing.
		if (node->nodeEnv->invariableSites) {
			node->nodeEnv->invariableSites = false;
			node->invariable.clear();
			node->nodeEnv->proportion_invariable = 0;
		}
		switch (node->nodeEnv->rateHetero) {
		case GammaRates:
			node->gammaRates.clear();
			node->nodeEnv->gammaShape = 1;
			break;
		case DiscreteGammaRates:
			node->categories.clear();
			node->nodeEnv->gammaShape = 1;
			break;
		default:
			break;
		}
		node->nodeEnv->rateHetero = NoRates;
    } else if (node->nodeEnv->invariableSites) {
		Create_invarRates(node, inNumSites); 
        for(int i=0; i<inNumSites; i++) {
			if ( !(iTree->randomInvariableAssignment) ) 
				node->invariable.at(i) = node->anc->invariable.at(i);
			else if (node->anc->nodeEnv->invariableSites) {
				// Preserves invariable regions.
				double ratio, r;
				if (node->anc->nodeEnv->proportion_invariable < node->nodeEnv->proportion_invariable) {
					ratio = node->nodeEnv->proportion_invariable - node->anc->nodeEnv->proportion_invariable;
					if (node->anc->invariable.at(i) == NO_CONSTRAINT) {
						r = rndu();
						if (r < ratio) node->invariable.at(i) = INVARIABLE;
						else node->invariable.at(i) = NO_CONSTRAINT;
					} else {
						if (node->anc->invariable.at(i) == NO_INDEL || node->anc->invariable.at(i) == INVAR_AND_NOINDEL)
							node->invariable.at(i) = node->anc->invariable.at(i);
						else node->invariable.at(i) = INVARIABLE;
					}
				} else if (node->anc->nodeEnv->proportion_invariable > node->nodeEnv->proportion_invariable) {
					ratio = node->nodeEnv->proportion_invariable / node->anc->nodeEnv->proportion_invariable;
					if (node->anc->invariable.at(i) == INVARIABLE) {
						r = rndu();
						if (r < ratio) node->invariable.at(i) = NO_CONSTRAINT;
						else node->invariable.at(i) = INVARIABLE;
					} else node->invariable.at(i) = node->anc->invariable.at(i);
				} else node->invariable.at(i)=node->anc->invariable.at(i);
			} else {
				// No invariable sites, therefore set the proportion_invariable.
				node->invariable.at(i)=IsInvariable(node->nodeEnv->proportion_invariable);
			}
        }

		if ( !(iTree->randomInvariableAssignment) ) {
			int num_invar = 0;
			for (int i = 0; i < inNumSites; i++) {
				if (node->invariable.at(i) == INVARIABLE || node->invariable.at(i) == INVAR_AND_NOINDEL)	
					num_invar++;
			}
			node->nodeEnv->proportion_invariable = (double)num_invar / (double)inNumSites;
		}
    } 

// Is this now taken care of since invariable stes are explicitly represented in motifSites?
//	if (node->nodeEnv->invariableSites) *en_invar_scale = (1.0 - node->nodeEnv->proportion_invariable);

	if (node->nodeEnv->constraintChange != PSEUDOGENE)
		SetCategories(node, inNumSites, options);
}

void initializeTreeScalingParameters(inTree *iTree, double *scale, double *invar_scale) 
{
	// Calc scale
	*scale = iTree->partitionRate;
	if (iTree->scaleTrees) {
		if (!iTree->my_tree->rooted) {
			cerr << "To scale tree length, trees must be rooted and ultrametric." << endl;
			exit(EXIT_FAILURE);
		}
		*scale *= iTree->treeScale / iTree->my_tree->totalLength;
	} else if (iTree->scaleBranches) {
		*scale *= iTree->branchScale;
		cerr << "Scaling branches by factor of " << iTree->branchScale << endl;
	}

	// Calc invar_scale.
	if (iTree->my_tree->treeEnv.front()->invariableSites) {
		if (iTree->randomInvariableAssignment) {
			if (iTree->rootSeqType == RANDOM) {
				for (int j=0; j<iTree->partitionLength; j++) {
   					iTree->my_tree->root->invariable.at(j) = IsInvariable(iTree->proportion_invariable); 
				}
   				for (int j=iTree->partitionLength; j<iTree->partitionLength; j++) 
   					iTree->my_tree->root->invariable.at(j)=0;
			} else {
				cerr << "Didn't think I could get to this spot:: InitializeTreeParameters, iTree->randomInvariableAssignment, iTree->rootSeqType != RANDOM." << endl;
				exit(EXIT_FAILURE);
			}
		} else {
			if (!(iTree->rootSeqType == RANDOM)) {
				for (int j=0; j<iTree->partitionLength; j++) 
					if (iTree->my_tree->root->invariable.at(j) == 1 || iTree->my_tree->root->invariable.at(j) == 3)
						(iTree->proportion_invariable)++;
				iTree->proportion_invariable /= iTree->partitionLength;
			} else {
				cerr << "Didn't think I could get to this spot:: InitializeTreeParameters, !(iTree->randomInvariableAssignment), iTree->rootSeqType == RANDOM." << endl;
				exit(EXIT_FAILURE);
			}
		}
		*invar_scale = (1.0 - iTree->my_tree->treeEnv.front()->proportion_invariable);
	} else *invar_scale = 1;
}

void initializeMotifPositions(inTree *iTree, TNode *des, seqGenOptions *options, double subst_len, double *motif_scale) 
{
	size_t current_position;

	des->myMotif.clear();
	des->Inherit_varSites(des->anc->variable_region_list);

	if ( !(des->nodeEnv->constraintChange == PSEUDOGENE) )
		des->InheritMotifSites();
	else {
		des->clearMotifSites();
		return;
	}

	des->setInvariableArrayPos();

	// Inherited motifs
	current_position = 0;
	if (des->myMotif.size() > 1)
		for (vector<motifSite*>::iterator it = des->myMotif.begin()+1; it != des->myMotif.end(); it++, current_position++)
			*motif_scale += CalculateMotifScale(des, current_position, subst_len, (*it)->activeProps->subst->substitution_bitstring);

	*motif_scale /= des->sequence.size();
}

double CalculateMotifScale(TNode *des, int position, double subst_len, bitset<20>& motif_characters)
{
	double scale = 0;
	double position_score, in_val;
	int cat;
	int num_pos = 0;

	switch (des->nodeEnv->rateHetero) {
	case GammaRates:
		for (int i = 0; i < numStates; i++) {
			position_score = 0;
			if (motif_characters.test(i)) {
				num_pos++;
				in_val = des->gammaRates.at(position)*subst_len;
				SetVector(cvector, i, in_val);
				for (int j = 0; j < numStates; j++)
					if (!motif_characters.test(j))
						scale += ( (j == (numStates-1)) ? (cvector[j+1]-cvector[j]) : (1-cvector[j]) );
			}
		}
	    break;

	case DiscreteGammaRates:
		cat = des->categories.at(position);
		SetMatrix(matrix[cat], des->nodeEnv->catRate[cat]*subst_len);
		for (int i = 0; i < numStates; i++) {
			position_score = 0;
			if (motif_characters.test(i)) {
				num_pos++;
				for (int j = 0; j < numStates; j++)
					if (!motif_characters.test(j))
						scale += ( (j == (numStates-1)) ? (1-matrix[cat][i*numStates+j]) : (matrix[cat][i*numStates+j+1]-matrix[cat][i*numStates+j]) );
			}
		}
	    break;

	case CodonRates:
		cat = position % 3;
		SetMatrix(matrix[cat], des->nodeEnv->catRate[cat]*subst_len);
		for (int i = 0; i < numStates; i++) {
			position_score = 0;
			if (motif_characters.test(i)) {
				num_pos++;
				for (int j = 0; j < numStates; j++)
					if (!motif_characters.test(j))
						scale += ( (j == (numStates-1)) ? (1-matrix[cat][i*numStates+j]) : (matrix[cat][i*numStates+j+1]-matrix[cat][i*numStates+j]) );
			}
		}
	    break;

	case NoRates:
	    SetMatrix(matrix[0], subst_len);
		for (int i = 0; i < numStates; i++) {
			position_score = 0;
			if (motif_characters.test(i)) {
				num_pos++;
				for (int j = 0; j < numStates; j++)
					if (!motif_characters.test(j))
						scale += ( (j == (numStates-1)) ? (1-matrix[0][i*numStates+j]) : (matrix[0][i*numStates+j+1]-matrix[0][i*numStates+j]) );
			}
		}
	    break;
	} 

	return scale / (double)num_pos;
}

int testRelation(vector<bool>& clade, vector<bool>& node)
{
	bool all_bits_same = true;
	bool some_bits_same = false;

	if(clade.size() != node.size()) {
		cerr << "Error: Bipartitions in testRelation are not the same size!?!?" << endl;
		exit(EXIT_FAILURE);
	}
	vector<bool>::iterator it2 = node.begin();
	for (vector<bool>::iterator it = clade.begin(); it != clade.end(); it++, it2++) {
		if ( (*it) != (*it2) ) all_bits_same = false;
		if ( (*it) && (*it2) ) some_bits_same = true;
	}
	if (all_bits_same) return EXACT;

	if (some_bits_same) {
		vector<bool>::iterator it2 = node.begin();
		for (vector<bool>::iterator it = clade.begin(); it != clade.end(); it++, it2++) {
			if ( (*it) && !(*it2) ) return NODE_IS_DES;
			if ( !(*it) && (*it2) ) return NODE_IS_ANC;
		}
	} else return NO_RELATION;
}

void EvolveNode(inTree *iTree, TNode *anc, TNode *des, double scale, double invar_scale, int inNumSites, int *P, list<eventTrack*> *events, seqGenOptions *options)
{ 
    double subst_len, indel_len;   
    double en_scale = scale, en_invar_scale = invar_scale;
	double motif_scale = 0;

	ofstream step_out;
	string trs_outfile;

	node_num++;
	if (!options->quiet && node_num % 10 == 0) cerr << "Node " << node_num << endl;
	if (output_trs_example) {
		for (vector<bool>::iterator it = des->bipartition.begin(); it != des->bipartition.end(); it++) {
			if ( (*it) ) trs_outfile += "1";
			else trs_outfile += "0";
		}
		cout << trs_outfile;
		trs_outfile += ( (options->simulation_step_type == DISCRETE_EVOLUTIONARY_STEPS) ? ".des" : ".trs");
		step_out.open(trs_outfile.c_str(), ios::trunc | ios::out);
	}

    des->anc=anc;
	des->sequence.assign(anc->sequence);

    *P+=1;
    des->mytipNo=*P;
	des->atEpochTime = anc->atEpochTime;
	des->example_epoch_time = anc->example_epoch_time;

	initializeCladeParameters(iTree->my_tree, des);
	initializeBranchRun(iTree, des, &en_invar_scale, anc->sequence.size(), options);

	//des->Print_Active_Sites(des->output_sequence(des->sequence.begin(), des->sequence.end()));

	if (options->simulation_step_type == INDEL_SEQ_GEN_V1) {
		des->numEventsToSimulate = 1;
		subst_len = des->length0 * (en_scale / en_invar_scale);
		indel_len = des->length0 * en_scale;
	} else {
	    subst_len=des->BL_step_size * (en_scale / en_invar_scale);
		indel_len=des->BL_step_size * en_scale;
	}

	initializeMotifPositions(iTree, des, options, subst_len, &motif_scale);
	motif_scale /= des->numEventsToSimulate;

	if (output_trs_example) {
		if ( options->simulation_step_type == DISCRETE_EVOLUTIONARY_STEPS ) {
			step_out << des->example_epoch_time << endl;
			cout << ": " << des->BL_step_size << endl;
		} else if (options->simulation_step_type == TIME_RELATIVE_STEPS) { 
			step_out << des->atEpochTime << endl;
			cout << ": " << des->BL_step_size << endl;
		}
	}

	for (int i = 0; i < des->numEventsToSimulate; i++) {
		des->atEpochTime += iTree->my_tree->epoch_step_size;
		if (output_trs_example) {
			des->example_epoch_time += iTree->my_tree->trs_example_step_size;
			if ( options->simulation_step_type == DISCRETE_EVOLUTIONARY_STEPS )
				step_out << des->example_epoch_time << endl;
			else step_out << des->atEpochTime << endl;
		}
    	MutateSequence(iTree, des, des->sequence, subst_len, motif_scale, events);
    	if (des->nodeEnv->indelFlag)
	   		InDel(iTree->my_tree,des,indel_len,events, des->numEventsToSimulate);
	}

	if (output_trs_example) step_out.close();

    if (des->tipNo==-1) { 
        EvolveNode(iTree, des, des->branch1, scale, invar_scale, inNumSites, P, events, options);
    	EvolveNode(iTree, des, des->branch2, scale, invar_scale, inNumSites, P, events, options);
    } else des->calcMotifAccuracy();	// calculator for pub (both iSGv1 & iSGv2)

	des->Remove_Objects();
} 
                 
void EvolveSequences(inTree *iTree, list<eventTrack*> *events, seqGenOptions *options)
{
    int *node_num;
	double scale, invar_scale;
	num_subst = 0;
    node_num= (int*)malloc(sizeof(int));
    *node_num= -1;
	iTree->my_tree->root->mytipNo = *node_num;

	initializeTreeScalingParameters(iTree, &scale, &invar_scale);
	if (options->codon_only_indels) iTree->my_tree->codons_only = true;

	if (PRINT_ID) { 
		cout << ">" << endl; 
		num_insert = 0; 
		num_delete = 0;
		len_insert = 0;
		len_delete = 0;
	}
	
	EvolveNode(iTree, iTree->my_tree->root, iTree->my_tree->root->branch1, scale, invar_scale, iTree->partitionLength, node_num, events, options);
    EvolveNode(iTree, iTree->my_tree->root, iTree->my_tree->root->branch2, scale, invar_scale, iTree->partitionLength, node_num, events, options);
    if (!iTree->my_tree->rooted)
        EvolveNode(iTree, iTree->my_tree->root, iTree->my_tree->root->branch0, scale, invar_scale, iTree->partitionLength, node_num, events, options);

	if (iTree->rootSeqType != RANDOM) iTree->my_tree->root->Remove_Objects(true);
	else {
		for (vector<motifSite*>::iterator it = iTree->my_tree->root->myMotif.begin(); it != iTree->my_tree->root->myMotif.end(); it++) {
			for (list<siteProperties*>::iterator it2 = (*it)->site_props.begin(); it2 != (*it)->site_props.end(); it2++) {
				delete (*it2)->indel->del;
				delete (*it2)->indel->L_ins_;
				delete (*it2)->indel->R_ins_;
				delete (*it2)->indel;
				delete (*it2)->subst;
				delete (*it2);
			}
			(*it)->site_props.clear();
			delete (*it);
		}
		iTree->my_tree->root->myMotif.clear();

		for (list<varSite*>::iterator it = iTree->my_tree->root->variable_region_list.begin(); it != iTree->my_tree->root->variable_region_list.end(); it++) {
			delete (*it);
		}

	}
	free(node_num);
} 

void Print_Globals(TTree *tree) 
{ 
    int i; 
    stringstream getlen;
    int fromAnc_len, indelNo_len;
   
    getlen.str("");
	// action
    for(i=1; i<tree->global_alignment_action.size(); i++) {
        if(tree->global_alignment_fromAnc.at(i) == -1) { fprintf(stderr,"-"); }
        else {
			if(tree->global_alignment_fromAnc.at(i) != -1) {
        		getlen << tree->global_alignment_fromAnc.at(i);
        		fromAnc_len = (getlen.str()).size();
        		getlen.str("");
        	} else fromAnc_len = 1;
        	if(tree->global_alignment_indelNo.at(i) != -1) {
        		getlen << tree->global_alignment_indelNo.at(i);
        		indelNo_len = (getlen.str()).size();
				getlen.str("");
			} else indelNo_len = 1;

			int numspace = (fromAnc_len > indelNo_len) ? fromAnc_len : indelNo_len;
			if(numspace > 1) 
				for(int j = 1; j < numspace; j++) fprintf(stderr, " ");

            fprintf(stderr,"%c",tree->global_alignment_action.at(i));
        }    
    } 

	// fromAnc
    fprintf(stderr,"\n"); 
    for(i=1; i<tree->global_alignment_action.size(); i++) {
        if(tree->global_alignment_fromAnc.at(i) == -1) { fprintf(stderr,"-"); }
        else { 
        	if(tree->global_alignment_indelNo.at(i) != -1) {
        		getlen << tree->global_alignment_indelNo.at(i);
        		indelNo_len = (getlen.str()).size();
        		getlen.str("");
			} else indelNo_len = 1;

			if(indelNo_len > 1) 
				for(int j = 1; j < indelNo_len; j++) fprintf(stderr, " ");

        	fprintf(stderr,"%d",tree->global_alignment_fromAnc.at(i)); 
        }
    }
    fprintf(stderr,"\n");

	// indelNo
	for(i=1; i<tree->global_alignment_action.size(); i++) {
        if(tree->global_alignment_indelNo.at(i) == -1) { fprintf(stderr,"-"); }
        else { 
        	if(tree->global_alignment_fromAnc.at(i) != -1) {
        		getlen << tree->global_alignment_fromAnc.at(i);
        		fromAnc_len = (getlen.str()).size();
        		getlen.str("");
			} else fromAnc_len = 1;

			if(fromAnc_len > 1) 
				for(int j = 1; j < fromAnc_len; j++) fprintf(stderr, " ");

        	fprintf(stderr,"%d",tree->global_alignment_indelNo.at(i)); 
        }
    }
	fprintf(stderr,"\n");
} 

