// Copyright (C) 2008 Cory Strope <corystrope@gmail.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdio.h>
#include <stdlib.h>
#include <cstring>
#include <math.h>
#include <ctype.h>
#include <string>
#include <list>
#include <iostream>
#include <iomanip>

#include "treefile.h"
#include "twister.h"
#include "evolve.h"

#define INSERT 1
#define DELETE 2

char treeErrorMsg[256];
int treeError;
extern int run_no;

TNode *avail=NULL;
long usedAvail=0;
long usedMalloc=0;

/* functions */

TNode::TNode(TTree *tree)
{
	branch0=branch1=branch2=NULL;
//	codons_only = false;
	length0=length1=length2=0.0;
	param=0.0;
	tipNo=-1;
	mytipNo=-1;
	anc=NULL;
	sequence.clear();
	gammaRates.clear();
	invariable.clear();
	categories.clear();
	clade_label.clear();
	tree->numNodes++;
	indel_num = 0;
	bipartition.clear();
	DistanceFromRoot = 0.0;
	atEpochTime = 0.0;
	example_epoch_time = 0.0;
	branch1_max_path = branch2_max_path = 0.0;
	branch1_time_relative_length=branch2_time_relative_length=branch0_time_relative_length=0.0;
	nodeEnv=NULL;
	addGeneral_varSites();
} 

void TNode::addGeneral_varSites()
{
	variable_region_list.clear();
	one_site_varSite = new varSite(1,1,this);
	unconstrained_varSite = new varSite(0,10000,this);
}

void TNode::report_varSites()
{
	size_t total_varSites = 0;

	cout << "Address\tminl\tmaxl\tcurrent_num_members " << endl;
	for (list<varSite*>::iterator it = variable_region_list.begin(); it != variable_region_list.end(); it++) {
		cout << (*it) << "\t";
		cout << (*it)->min << "\t" << (*it)->max << "\t" << (*it)->members.size() << endl;
		total_varSites += (*it)->members.size();
	}

	cout << "The total number of varSites is: " << total_varSites << endl;
}

void TNode::report_on_sites()
{
	for (vector<motifSite*>::iterator it = myMotif.begin(); it != myMotif.end(); it++) {
		cerr << "L_ins_ (st=" << (*it)->activeProps->indel->L_ins_->on_site_sequence_template_varSite
		     << " m=" << (*it)->activeProps->indel->L_ins_->on_site_motif_varSite << ")   ";
		cerr << "R_ins_ (st=" << (*it)->activeProps->indel->R_ins_->on_site_sequence_template_varSite
		     << " m=" << (*it)->activeProps->indel->R_ins_->on_site_motif_varSite << ") " << endl;
	}

}

varSite *TNode::Generic_varSite()
{
	return one_site_varSite;
}

void TNode::report(TTree *tree) 
{
	for (vector<bool>::iterator it = bipartition.begin(); it != bipartition.end(); it++)
		cerr << (*it);

	if (tipNo != -1 && tree != NULL)
		cerr << " \"" << tree->names.at(tipNo) << "\"";
}

void TNode::Inherit_varSites(list<varSite*> anc_varList) 
{
	size_t i = 0;

	anc->one_site_varSite->descendant_equiv = one_site_varSite;
	anc->unconstrained_varSite->descendant_equiv = unconstrained_varSite;
	for (list<varSite*>::iterator it = anc_varList.begin(); it != anc_varList.end(); it++, i++)
		if (i > 1) variable_region_list.push_back(new varSite(*it,this));

	list<varSite*>::iterator anc_it = anc_varList.begin();
	for (list<varSite*>::iterator it = variable_region_list.begin(); it != variable_region_list.end(); it++, anc_it++) {
		if (anc_it == anc_varList.end()) {
			cerr << "something's wrong" << endl;
			exit(EXIT_FAILURE);
		}
	}
}

void TNode::InheritMotifSites()
{	
	// Create new positions for each. Setting active sites will copy new things into each pos.
	size_t i = 0;
	for (vector<motifSite*>::iterator it = anc->myMotif.begin(); it != anc->myMotif.end(); it++) {
		myMotif.push_back(new motifSite(*it));
	}
	this->postProcess();

	bool last_site = false;
	myMotif.front()->activeProps = new siteProperties();
	vector<motifSite*>::iterator it3 = anc->myMotif.begin()+1;
	for (vector<motifSite*>::iterator it2 = myMotif.begin()+1; it2 != myMotif.end(); it2++, it3++) {
		if ( it2==myMotif.end()-1 ) last_site = true;
		(*it2)->setActiveProps(bipartition, (*(it2-1)), (*it3), nodeEnv, last_site);
	}
	setRemainingPos();

//cerr << "Ancestor: " << endl;
//anc->Print_Active_Sites(output_sequence(anc->sequence.begin(), anc->sequence.end()));
//cerr << "Descendant: " << endl;
//	Print_Active_Sites(output_sequence(sequence.begin(), sequence.end()));
//exit(0);
	setZeroSites();
}

void TNode::setFirstMotifPos()
{
	vector<motifSite*>::iterator it = myMotif.begin();

	// Beginning should be a (1,1) varSite.
	(*it)->activeProps->indel->del->my_sequence_template_varSite 
	= (*it)->activeProps->indel->del->my_motif_varSite 
	= one_site_varSite;

	// Need to manually set the first position insertion sites, since the connecting routine
	// checks only those sites that 
	(*it)->activeProps->indel->L_ins_->my_sequence_template_varSite[LEFT]
	= (*it)->activeProps->indel->L_ins_->my_motif_varSite[LEFT]
	= one_site_varSite;
	(*it)->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT]
	= (*it)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]
	= one_site_varSite;

	(*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]
	= (*(it+1))->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT];
	(*it)->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]
	= (*(it+1))->activeProps->indel->L_ins_->my_motif_varSite[RIGHT];

}

void TNode::setRemainingPos()
{
	vector<motifSite*>::iterator it;

	setFirstMotifPos();

	for (list<varSite*>::iterator it2 = variable_region_list.begin(); it2 != variable_region_list.end(); it2++) {
		for (it = myMotif.begin(); it != myMotif.end(); it++) {
			// If any varSites remain NULL, they should be unconstrained.
			if ( (*it)->activeProps->indel->del->my_sequence_template_varSite == NULL) {
				(*it)->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT]
				= (*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]
				= (*it)->activeProps->indel->del->my_sequence_template_varSite 
				= unconstrained_varSite;
			}
			if ( (*it)->activeProps->indel->del->my_motif_varSite == NULL) {
				(*it)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]
				= (*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]
				= (*it)->activeProps->indel->del->my_motif_varSite 
				= unconstrained_varSite;
			}
		}
	}
	
	it = myMotif.begin();
	(*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]
	= (*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT] 
	= one_site_varSite;
	
	(*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]
	= (*(it+1))->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
	
	// Set each of the positions at the beginning and end of motif positions.
	for (it = myMotif.begin(); it != myMotif.end()-1; it++) {
		// Each of these will come in pairs, except for the first and last positions.
		if ( (*it)->activeProps->indel->R_ins_->my_motif_varSite[RIGHT] == NULL ) {
			if ((*(it+1))->activeProps->indel->del->my_motif_varSite == NULL) {
				cerr << "Please integrate motifs that are coincident to each other that activate on the same clade into a single motif." << endl;
				exit(EXIT_FAILURE);
			}
			(*it)->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]
			= (*(it+1))->activeProps->indel->L_ins_->my_motif_varSite[RIGHT] 
			= (*(it+1))->activeProps->indel->del->my_motif_varSite;
		}

		if ( (*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT] == NULL ) {
			if ( (*it)->activeProps->indel->del->my_motif_varSite == NULL ) {
				cerr << "Please integrate motifs that are coincident to each other that activate on the same clade into a single motif." << endl;
				exit(EXIT_FAILURE);
			}
			(*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]
			= (*(it+1))->activeProps->indel->L_ins_->my_motif_varSite[LEFT]
			= (*it)->activeProps->indel->del->my_motif_varSite;
		}
	}
	
	it = myMotif.end()-1;
	(*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT] = (*it)->activeProps->indel->del->my_sequence_template_varSite;
	(*it)->activeProps->indel->R_ins_->my_motif_varSite[RIGHT] = (*it)->activeProps->indel->del->my_motif_varSite;

	for (vector<motifSite*>::iterator it2 = myMotif.begin()+1; it2 != myMotif.end(); it2++) {
		(*it2)->activeProps->indel->L_ins_->my_sequence_template_varSite[LEFT] = (*(it2-1))->activeProps->indel->del->my_sequence_template_varSite;
		(*it2)->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT] = (*it2)->activeProps->indel->del->my_sequence_template_varSite;
		(*it2)->activeProps->indel->L_ins_->my_motif_varSite[LEFT] = (*(it2-1))->activeProps->indel->del->my_motif_varSite;
		(*it2)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT] = (*it2)->activeProps->indel->del->my_motif_varSite;
	}

	for (list<varSite*>::iterator it2 = variable_region_list.begin(); it2 != variable_region_list.end(); it2++) {
		for (it = myMotif.begin(); it != myMotif.end(); it++) {
			// Each site should belong to 2 varSites: one for template, one for motif.
			if ( (*it)->activeProps->indel->del->my_sequence_template_varSite == (*it2))
				(*it2)->add2Member((*it)->activeProps->indel->del);
			if ( (*it)->activeProps->indel->del->my_motif_varSite == (*it2))
				(*it2)->add2Member((*it)->activeProps->indel->del);
		}
	}
}

void TNode::setInvariableArrayPos()
{
	vector<motifSite*>::iterator it;

	// Nearly all set, but still have to represent the invariable array constraints, applies as motif.
	if (nodeEnv->invariableSites) { 
		vector<short>::iterator iit = invariable.begin()+1;
		string::iterator sit = sequence.begin()+1;
		it = myMotif.begin()+1;
		for ( ; it != myMotif.end(); it++, iit++, sit++) {
			bool last_site = ( (it == myMotif.end()-1) ? true : false );
			if ( !(*it)->activeProps->fromMotif[MOTIF] ) {
				// Need to make current position invariable, and change motif pointers of the
				// previous and next position (if they exist).
				switch ( (*iit) ) {
				case NO_CONSTRAINT:
					// Do nothing
					break;
				case INVARIABLE:
					// Set to (1,1), make subst equal to only current aa.
					(*it)->activeProps->indel->del->my_motif_varSite
					= (*it)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]
					= (*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]
					= one_site_varSite;
					
					(*it)->activeProps->indel->L_ins_->on_site_motif_varSite 
					= (*it)->activeProps->indel->R_ins_->on_site_motif_varSite 
					= unconstrained_varSite;

					(*it)->activeProps->subst->setInvariable(*sit);
					break;
				case NO_INDEL:
					// set to (1,1), check next positions of invar to see if R_ins_->on_site needs set to (1,1)
					(*it)->activeProps->indel->del->my_motif_varSite
					= (*it)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]
					= (*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]
					= one_site_varSite;

					if ( !last_site ) {
						if ( (*(iit+1)) == NO_INDEL || (*(iit+1)) == INVAR_AND_NOINDEL ) {
							(*it)->activeProps->indel->R_ins_->on_site_motif_varSite = one_site_varSite;
						}
					} // Else we don't set anything.
					break;
				case INVAR_AND_NOINDEL:
					// Do both INVAR & INDEL.
					(*it)->activeProps->indel->del->my_motif_varSite
					= (*it)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]
					= (*it)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]
					= one_site_varSite;
	
					if ( !last_site ) {
						if ( (*(iit+1)) == NO_INDEL || (*(iit+1)) == INVAR_AND_NOINDEL ) {
							(*it)->activeProps->indel->R_ins_->on_site_motif_varSite = one_site_varSite;
						}
					} // Else we don't set anything.	

					(*it)->activeProps->subst->setInvariable(*sit);
					break;
				default:
					cerr << "What kind of invariable array position is " << (*iit) << endl;
					exit(EXIT_FAILURE);
					break;
				}
			}
		}
	}
}

bool TNode::isVarSite(varSite *chk_varSite)
{
	if (chk_varSite == NULL) return true;

	for (list<varSite*>::iterator it = variable_region_list.begin(); it != variable_region_list.end(); it++)
		if (chk_varSite == (*it) ) return true;

	return false;
}

void TNode::setZeroSites()
{
	bool doit;
	size_t i = 0;

	myMotif.front()->activeProps->indel->L_ins_->on_site_motif_varSite
	= myMotif.front()->activeProps->indel->L_ins_->on_site_sequence_template_varSite
	= NULL;

	for (vector<motifSite*>::iterator it = myMotif.begin(); it != myMotif.end(); it++, i++) {
		if ( (*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->min == 0) {
			(*it)->activeProps->indel->R_ins_->on_site_sequence_template_varSite = (*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT];
		}
		else {
			(*it)->activeProps->indel->R_ins_->on_site_sequence_template_varSite = NULL;
		}

		if ( (*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->min == 0) {
			(*it)->activeProps->indel->R_ins_->on_site_sequence_template_varSite = (*it)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT];
		}
		else {
			(*it)->activeProps->indel->R_ins_->on_site_sequence_template_varSite = NULL;
		}

		doit = true;
		if ( (*it)->activeProps->indel->R_ins_->on_site_motif_varSite != NULL ) {
			if (isVarSite( (*it)->activeProps->indel->R_ins_->on_site_motif_varSite)) {
				if ( (*it)->activeProps->indel->R_ins_->on_site_motif_varSite->min == 0 ) {
					doit = false;
				}
			}  else (*it)->activeProps->indel->R_ins_->on_site_motif_varSite = NULL;
		}

		if ( it != myMotif.end()-1 && doit) {
			// If either flanking site has a zero site, our sequence must have it.
			try {
				//cerr << "position: " << i << endl;
				if ( (*it)->activeProps->indel->del->my_motif_varSite->min == 0 ) {
					(*it)->activeProps->indel->R_ins_->on_site_motif_varSite = (*it)->activeProps->indel->del->my_motif_varSite;
				} else if ( (*(it+1))->activeProps->indel->del->my_motif_varSite->min == 0 ) {
					(*it)->activeProps->indel->R_ins_->on_site_motif_varSite = (*(it+1))->activeProps->indel->del->my_motif_varSite;
				} else {
				}
			} catch (exception &e) {
				cerr << "Error in site " << i << ": " << e.what() << "." << endl;
				exit(EXIT_FAILURE);
			}
		}
	}
}

void TNode::FULL_REPORT()
{
	cout << "anc node: " << anc << endl;
	cout << "des node: " << this << endl;

	list<varSite*>::iterator anc_vS = anc->variable_region_list.begin();
	list<varSite*>::iterator des_vS = variable_region_list.begin();
	for (; des_vS != variable_region_list.end(); des_vS++, anc_vS++) {
		cout << setw(10) << (*anc_vS) << setw(4) << (*anc_vS)->min << setw(8) << (*anc_vS)->max << setw(10) << (*anc_vS)->descendant_equiv << setw(5) << (*anc_vS)->members.size();
		cout << setw(14) << "des " << (*des_vS) << setw(4) << (*des_vS)->min << setw(8) << (*des_vS)->max << setw(5) << (*des_vS)->members.size() << endl;
	}

	vector<motifSite*>::iterator anc_it = anc->myMotif.begin();
	vector<motifSite*>::iterator des_it = myMotif.begin();	
	for ( ; anc_it != anc->myMotif.end(); anc_it++, des_it++) {
		cout << "subst: ";
		cout << setw(10) << (*anc_it)->activeProps << setw(21) << (*anc_it)->activeProps->subst->report_bitset();
		cout << setw(10) << (*des_it)->activeProps << setw(21) << (*des_it)->activeProps->subst->report_bitset();
		cout << endl;
	}
}

void TNode::clearMotifSites() 
{
	size_t i;

	// Pseudogened. Make all positions (except for the first) an unconstrained site.
	// 1: create motifSite array.
	for (i = 0; i < anc->myMotif.size(); i++) {
		myMotif.push_back(new motifSite());
	}

	postProcess();
	myMotif.front()->activeProps = new siteProperties();

	// 2 manually set first position as a single site (non-deletable.)
	myMotif.front()->site_props.front()->indel->del->my_sequence_template_varSite 
	  = myMotif.front()->site_props.front()->indel->del->my_motif_varSite 
	  = one_site_varSite;
	bool last_site = false;
	for (vector<motifSite*>::iterator it2 = myMotif.begin()+1; it2 != myMotif.end(); it2++, i++) {
		if ( it2==myMotif.end()-1 ) last_site = true;
		(*it2)->setActiveProps(bipartition, (*(it2-1)), NULL, nodeEnv, last_site);
	}

	setRemainingPos();
	setZeroSites();
}

string TNode::output_sequence(string::iterator start, string::iterator end)
{
	string seqout = "";
	for (string::iterator it = start; it != end; it++) {
		seqout += stateCharacters[*it];
	}
	return seqout;
}

void TNode::Remove_Objects(bool root_node)
{
	bool aP;
	size_t i = 0;
	
	delete myMotif.back()->activeProps->indel->R_ins_;
	
	for (vector<motifSite*>::iterator it = myMotif.begin(); it != myMotif.end(); it++,i++) {
		aP = false;
		for (list<siteProperties*>::iterator it2 = (*it)->site_props.begin(); it2 != (*it)->site_props.end(); it2++) {
			// General sites become activeProps. If I run across a general site, I don't want to
			// delete the activeProps (b/c much non-hilarity ensues...)
			if ( (*it2)->indel->L_ins_ != (*it)->activeProps->indel->L_ins_) {
				if ( it2 == (*it)->site_props.begin() ) {
					delete (*it2)->indel->L_ins_;
				} else {
					if ( !root_node )
						delete (*it2)->indel->L_ins_;
				}
			} else aP = true;
			if ( (*it2)->indel->R_ins_ != (*it)->activeProps->indel->R_ins_) 
				delete (*it2)->indel->R_ins_;
			if ( (*it2)->indel->del != (*it)->activeProps->indel->del)
				delete (*it2)->indel->del;
			if ( (*it2)->subst != (*it)->activeProps->subst)
				delete (*it2)->subst;
			if ( (*it2)->indel != (*it)->activeProps->indel)
				delete (*it2)->indel;
			if ( (*it2) != (*it)->activeProps)
				delete (*it2);
		}

		if ( !aP ) {
			delete (*it)->activeProps->indel->L_ins_;
			delete (*it)->activeProps->indel->del;
			delete (*it)->activeProps->subst;
			delete (*it)->activeProps->indel;
			delete (*it)->activeProps;
		}
		(*it)->site_props.clear();
		delete (*it);
	}
	myMotif.clear();

	// variable sites list.
	for (list<varSite*>::iterator it = variable_region_list.begin(); it != variable_region_list.end(); it++) {
		delete (*it);
	}
	variable_region_list.clear();	
}

void TTree::constructRootMotifSites(inTree *tree, string &root_sequence, string &invariable, int type) 
{
	int relationship;
	size_t motif_start, motif_end, found;
	vector<siteProperties*> site_properties;
	string::iterator sit, sit2;
	vector<motifSite*>::iterator mSit;
	vector<short> dot_additions2template (root_sequence.size(),0);
	vector<short>::iterator rm;
	
	// If any '.'s are placed on HMM-like match states, they will get here. This'll blast 'em.
	if (type != RANDOM) { 
		for (list<inClade*>::iterator it = treeEnv.begin(); it != treeEnv.end(); it++) {
			for (list<inMotif*>::iterator it2 = (*it)->my_motifs.begin(); it2 != (*it)->my_motifs.end(); it2++) {
				found = (*it2)->sitemap.find_first_of(".");
				while (found != string::npos) {
					for (list<inClade*>::iterator it4 = treeEnv.begin(); it4 != treeEnv.end(); it4++) {
						for (list<inMotif*>::iterator it3 = (*it4)->my_motifs.begin(); it3 != (*it4)->my_motifs.end(); it3++) {
							rm = dot_additions2template.begin()+found;
							if ( (*it3)->sitemap.at(found) == '*' || (*it3)->sitemap.at(found) == '.') {
								(*it3)->sitemap.erase(found,1);
								if ( (*it3)->isTemplate() ) {
									dot_additions2template.erase(rm);
								}
							} else {
								if ( (*it3)->isTemplate() ) {
									(*it3)->sitemap.erase(found,1);
									dot_additions2template.erase(rm);
									dot_additions2template.at(found) = 1;
								} else {
									cerr << "Cannot have a '.' in the motif specifications overlapping with another motif (template overlap is acceptable)" << endl;
									exit(EXIT_FAILURE);
								}
							}
						}
					}
					root_sequence.erase(found,1);
					invariable.erase(found,1);
					found = (*it2)->sitemap.find_first_of(".");
				}
			}
		}
	}

	// Allocate array
	root->myMotif.clear();
	for (int i = 0; i < root_sequence.size(); i++)
		root->myMotif.push_back(new motifSite());

	if (type == RANDOM) return;

	// Step 1: Relate all motifs to the root sequence.
	size_t dot_additions_sum = 0;
	for (vector<short>::iterator it = dot_additions2template.begin(); it != dot_additions2template.end(); it++) 
		dot_additions_sum += (*it);

	size_t motif_start_position, post_motif_size, post_motif, shouldbezero;
	for (list<inClade*>::iterator it = treeEnv.begin(); it != treeEnv.end(); it++) {
		for (list<inMotif*>::iterator it2 = (*it)->my_motifs.begin(); it2 != (*it)->my_motifs.end(); it2++) {
			relationship = testRelation( (*it)->bipartition, root->bipartition);
			site_properties = (*it2)->enumerateRegEx(root);
			motif_start_position = (*it2)->sitemap.find_first_not_of("*");
			post_motif = (*it2)->sitemap.find_first_of("*", motif_start_position);
			if (post_motif == string::npos) post_motif = root_sequence.size();			
			shouldbezero = root_sequence.size() - (site_properties.size() + motif_start_position + (root_sequence.size()-post_motif));
			if ( (*it2)->isTemplate() ) shouldbezero += dot_additions_sum;
			if (shouldbezero != 0) {
				cerr << "Motif \'" << (*it2)->marker << "': Regular expression size (" << site_properties.size() << ") does not agree with motif size (" << post_motif-motif_start_position << ")." << endl;
				exit(EXIT_FAILURE);
			}

			if ( (*it2)->isTemplate() ) {
				vector<siteProperties*>::iterator sPit = site_properties.begin();
				for (rm = dot_additions2template.begin(); rm != dot_additions2template.end(); rm++) {
					if ( (*rm) ) {
						site_properties.erase(sPit);
					}
					sPit++;
				}
			}

			// Set the vector to the motif site we are updating.
			mSit = root->myMotif.begin() + motif_start_position;
			sit = root_sequence.begin() + motif_start_position;
			rm = dot_additions2template.begin();

			// Get all of the siteProps to match the motifSites. Set active props later.
			for (vector<siteProperties*>::iterator it3 = site_properties.begin(); it3 != site_properties.end(); it3++) {
				if (relationship == EXACT) {
					if ( (*it2)->isTemplate() && (*rm) ) {
						// This is the case that the motif has '.', but template does not.
						(*mSit)->site_props.push_back(*it3);
					} else {
						(*mSit)->site_props.push_back(*it3);
					}
				} else if (relationship == NODE_IS_ANC) {
					// Push back property to the site prop. Only set dormant if not already an active site.
					(*mSit)->site_props.push_back(*it3);
					(*mSit)->site_props.back()->ancestral = false/*true*/;
				} else {
					cerr << "Root node can be nothing but EXACT or NODE_IS_ANC" << endl;
					exit(EXIT_FAILURE);
				}
				mSit++, rm++;
			}
		}
	}


	root->postProcess();
	root->myMotif.front()->activeProps = new siteProperties();

	bool last_site = false;
	sit = root_sequence.begin()+1;
	for (vector<motifSite*>::iterator it2 = root->myMotif.begin()+1; it2 != root->myMotif.end(); it2++) {
		if ( it2==root->myMotif.end()-1 ) last_site = true;
		(*it2)->setActiveProps(root->bipartition, (*(it2-1)), NULL, treeEnv.front(), last_site);
	}

	root->setRemainingPos();
	root->setZeroSites();

	// Remove sites that are '-', populate varSite list of pointers to motifSites that belong to
	// it. This will effectively remove the necessity of having the relateMotifSites function.
	found = root_sequence.rfind("-");
	while (found != string::npos) {
		for (list<inMotif*>::iterator it2 = tree->motif_specs.begin(); it2 != tree->motif_specs.end(); it2++)
			(*it2)->sitemap.erase(found,1);
		found = root_sequence.rfind("-",found-1);
	}

	sit = root_sequence.begin();
	sit2 = invariable.begin();
	int site_number = 0;
	for (vector<motifSite*>::iterator it = root->myMotif.begin(); it != root->myMotif.end(); it++) {
		// If the position is a '-', it (1) should be in a varSite for ALL siteProps, and (2) should
		// be removed from the sequence, also requiring the shifting of pointers. Otherwise,
		// No worries.
//		cerr << site_number << " " << (*sit) << endl;
		if ( (*sit) == '-') {
			(*it)->deleteMerge( (*(it-1)), (*(it+1)) );
			// Finish off by erasing this position.
			root->myMotif.erase(it);
			root_sequence.erase(sit);
			invariable.erase(sit2);
			it--, sit--, sit2--;
		} else site_number++;
		sit++, sit2++;
	}
	
//	cerr << endl << endl << "ROOT" << endl << endl;
//	root->Print_Active_Sites(root_sequence);
//	cerr << endl << endl << "ENDROOT" << endl << endl;
//	exit(0);
}

void TNode::postProcess()
{
	int i = 0;

	// Set all pointers for the motifSite array and deeper structures.
	for (vector<motifSite*>::iterator it = myMotif.begin(); it != myMotif.end(); it++, i++)
		(*it)->setNode(this);
	
	myMotif.front()->site_props.front()->indel->del->my_sequence_template_varSite 
	  = myMotif.front()->site_props.front()->indel->del->my_motif_varSite 
	  = one_site_varSite;
}

void TNode::dumpBetweenSitePointers(string root_sequence)
{
	string::iterator sit;
	int i = 0;

	sit = root_sequence.begin();
	for (vector<motifSite*>::iterator it2 = myMotif.begin(); it2 != myMotif.end(); it2++, i++) {
		// L_ins_ dump:
		if ( (*it2)->activeProps->indel->L_ins_ != NULL ) {
			cerr << "L_ins_: " << (*it2)->activeProps->indel->L_ins_;
			cerr << " st("<< (*it2)->activeProps->indel->L_ins_->my_sequence_template_varSite[LEFT]->min << "," << (*it2)->activeProps->indel->L_ins_->my_sequence_template_varSite[LEFT]->max << "**"
			     << (*it2)->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT]->min << "," << (*it2)->activeProps->indel->L_ins_->my_sequence_template_varSite[RIGHT]->max << ")\t";
			cerr << " m("<< (*it2)->activeProps->indel->L_ins_->my_motif_varSite[LEFT]->min << "," << (*it2)->activeProps->indel->L_ins_->my_motif_varSite[LEFT]->max << "**"
			     << (*it2)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]->min << "," << (*it2)->activeProps->indel->L_ins_->my_motif_varSite[RIGHT]->max << ")" << endl;
		} else cerr << "L_ins_ NULL" << endl;
			
		cerr << "\t" << i << " " << (*sit) << " " << (*it2) << " ";
		cerr << " st("<< (*it2)->activeProps->indel->del->my_sequence_template_varSite->min << "," << (*it2)->activeProps->indel->del->my_sequence_template_varSite->max << ")\t";
		cerr << " m("<< (*it2)->activeProps->indel->del->my_motif_varSite->min << "," << (*it2)->activeProps->indel->del->my_motif_varSite->max << ")" << endl;

		if ( (*it2)->activeProps->indel->R_ins_ != NULL ) {
			cerr << "R_ins_: " << setw(10) << (*it2)->activeProps->indel->R_ins_;
			cerr << " st("<< (*it2)->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->min << "," << (*it2)->activeProps->indel->R_ins_->my_sequence_template_varSite[LEFT]->max << "**"
			     << (*it2)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->min << "," << (*it2)->activeProps->indel->R_ins_->my_sequence_template_varSite[RIGHT]->max << ")\t";
			cerr << " m("<< (*it2)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->min << "," << (*it2)->activeProps->indel->R_ins_->my_motif_varSite[LEFT]->max << "**"
			     << (*it2)->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]->min << "," << (*it2)->activeProps->indel->R_ins_->my_motif_varSite[RIGHT]->max << ")" << endl;
		} else cerr << "R_ins_ NULL" << endl;

		cerr << endl;
		sit++;
	}
}

void TNode::Print_Active_Sites(string root_sequence)
{
	string::iterator sit;
	int i = 0;

	report(); cerr << endl;

	sit = root_sequence.begin();
	for (vector<motifSite*>::iterator it2 = myMotif.begin(); it2 != myMotif.end(); it2++, i++) {
		cerr << i << " " << (*sit) << " " << (*it2) << " ";
		if ( (*it2)->activeProps->fromMotif[TEMPLATE] != NULL) {
			cerr << "st active.   " << (*it2)->activeProps->indel->del->my_sequence_template_varSite->min << " " << (*it2)->activeProps->indel->del->my_sequence_template_varSite->max << "\t";
		} else cerr << "NONE     ";

		if ( (*it2)->activeProps->fromMotif[MOTIF] != NULL) {
			cerr << "m active.  " << (*it2)->activeProps->indel->del->my_motif_varSite->min << " " << (*it2)->activeProps->indel->del->my_motif_varSite->max << "\t";
		} else cerr << "NONE   ";

		if ( (*it2)->activeProps->indel->L_ins_ != NULL )
			cerr << "L_ins_ " << (*it2)->activeProps->indel->L_ins_ << "  ";
		if ( (*it2)->activeProps->indel->R_ins_ != NULL )
			cerr << "R_ins_ " << (*it2)->activeProps->indel->R_ins_ << "  ";
			
		if ( (*it2)->activeProps->indel->L_ins_->on_site_sequence_template_varSite != NULL && 
			 isVarSite((*it2)->activeProps->indel->L_ins_->on_site_sequence_template_varSite) )
			cerr << "Losst ";
		if ( (*it2)->activeProps->indel->L_ins_->on_site_motif_varSite != NULL && 
			 isVarSite((*it2)->activeProps->indel->L_ins_->on_site_motif_varSite) )
			cerr << "Losm(" << (*it2)->activeProps->indel->L_ins_->on_site_motif_varSite->min << " " << (*it2)->activeProps->indel->L_ins_->on_site_motif_varSite->max << ")";
		if ( (*it2)->activeProps->indel->R_ins_->on_site_sequence_template_varSite != NULL && 
			 isVarSite((*it2)->activeProps->indel->R_ins_->on_site_sequence_template_varSite) )
			cerr << "Rosst ";
		if ( (*it2)->activeProps->indel->R_ins_->on_site_motif_varSite != NULL && 
			 isVarSite((*it2)->activeProps->indel->R_ins_->on_site_motif_varSite) )
			cerr << "Rosm(" << (*it2)->activeProps->indel->R_ins_->on_site_motif_varSite->min << " " << (*it2)->activeProps->indel->R_ins_->on_site_motif_varSite->max << ")";

		cerr << endl;
		sit++;
	}
}

void TNode::calcMotifAccuracy()
{
	int motifActive;
	string::iterator sit = sequence.begin();
	for (vector<motifSite*>::iterator it = myMotif.begin(); it != myMotif.end(); it++,sit++) {
		if ( (*it)->activeProps->fromMotif[MOTIF] != NULL ) {
			total_motif_positions++;
			if ( (*it)->activeProps->subst->substitution_bitstring.test(*sit) ) {
				correct_motif_positions++;
			}		
		}
	}
}

void TTree::InitTree()
{	
	treeEnv.clear();
	root=NULL;
	nodeList.clear(); numNodes=0;
	names.clear();
	numTips=0;
	totalLength=0.0;
	rooted=0;
	lengths=-1;
	global_alignment_fromAnc.clear();
	global_alignment_indelNo.clear();
	global_alignment_action.clear();
}

TTree::TTree()
{
	InitTree();
}

TNode *TTree::AddNode() 
{
	TNode *node = new TNode(this);
	nodeList.push_back(node);
	
	return node;
}

inClade *TTree::AddClade(string& label) 
{
	inClade *clade = new inClade(label, this);
	treeEnv.push_back(clade);

	return clade;
}

TNode *TTree::ReadTip(string tree_str, int *pos, char ch, TTree *tree, bool first_partition)
{
	string name;
	size_t found;
	
	TNode *node = AddNode();
	
	found = tree_str.find_first_of(":,)",*pos);
	name = tree_str.substr(*pos, found - *pos);

	if(found - *pos > MAX_NAME_LEN) {
		cerr << "Name '" << name << "' longer than maximum allowed length." << endl;
		exit(EXIT_FAILURE);
	}

	*pos=found;	

	if (first_partition) {
		node->tipNo=tree->numTips;
		names.push_back(name);
	} else {
		size_t i = 0;
		while (i < names.size() && name.compare(names.at(i)) != 0)
			i++;
		if (i == names.size()) {
			sprintf(treeErrorMsg, "Taxon names in trees for different partitions do not match.");
			return NULL;
		}
		node->tipNo=i;
	}

	tree->tips.push_back(node);
	tree->numTips++;

	found = tree_str.find_first_of(":,)",*pos);
	*pos = found-1;

	if (found == tree_str.npos) {
		sprintf(treeErrorMsg, "Unexpected end of file");
		return NULL;
	}

	return node;
}

TNode *TTree::ReadNode(string tree_str, int *pos, TTree *tree, int detectPolytomies, bool first_partition)
{
	TNode *node, *node2;
	char ch;
	int found;

	try {
		node = AddNode();
	} catch (exception& e) {
		cerr << "Error creating TNode node (ReadTree): " << e.what() << endl;
		exit(EXIT_FAILURE);
	}

	if ((node2=ReadBranch(tree_str, pos, tree, first_partition))==NULL)
		return NULL;

	node->branch1=node2;
	node2->branch0=node;
	node->length1=node2->length0;

	ReadUntil(tree_str, pos, ',', "Comma");

	if (treeError)
		return NULL;
	if ((node2=ReadBranch(tree_str, pos, tree, first_partition))==NULL)
		return NULL;
	node->branch2=node2;
	node2->branch0=node;
	node->length2=node2->length0;
	
	found = tree_str.find_first_of(":,);", ++(*pos));
	*pos=found;
	ch = tree_str.at(found);
	
	if (detectPolytomies && ch==',') {
		fprintf(stderr, "This tree contains nodes which aren't bifurcations. Resolve the node\n");
		fprintf(stderr, "with zero branch lengths to obtain correct results. This can be done\n");
		fprintf(stderr, "with a program called TreeEdit: http://evolve.zoo.ox.ac.uk/software/TreeEdit\n");
		exit(EXIT_FAILURE);
	}

	if ((size_t)*pos == tree_str.npos) {
		sprintf(treeErrorMsg, "Unexpected end of file");
		return NULL;
	}
	(*pos)--;
	
	return node;
}

TNode *TTree::ReadBranch(string tree_str, int *pos, TTree *tree, bool first_partition)
{

	char ch;
	double len, param=0.0;
	TNode *node;
	int found;
	inClade *clade;

	ch=ReadToNextChar(tree_str, pos);

	if (ch=='(') {	// is a node
		node=ReadNode(tree_str, pos, tree, 1, first_partition);
		ReadUntil(tree_str, pos, ')', "Closing bracket");
		if (treeError)
			return NULL;
	} else {		// is a tip
		node=ReadTip(tree_str, pos, ch, tree, first_partition);
	}
	
	ch=ReadToNextChar(tree_str, pos);

	if (isalpha(ch)) {	// Specifying clade:
		string clade_label = "";
		while(isalpha(ch) || isdigit(ch) || ch == '_') {
			clade_label += ch;
			ch=ReadToNextChar(tree_str,pos);
		}
		clade = AddClade(clade_label);
		if(clade_label.size() <= MAX_NAME_LEN) {
			node->clade_label.assign(clade_label);
		} else {
			cerr << "Clade name " << clade_label << " has too many characters. Please shorten label." << endl;
			exit(EXIT_FAILURE);
		}
	}
	
	if (ch==':') {
		if (tree->lengths==0) {
			sprintf(treeErrorMsg, "Some branches don't have branch lengths");
			return NULL;
		} else 
			tree->lengths=1;

		if (sscanf(&tree_str.at(++(*pos)), "%lf", &len)!=1) {
			sprintf(treeErrorMsg, "Unable to read branch length");
			return NULL;
		}

		found = tree_str.find_first_not_of(".0192837465",*pos);
		*pos = found-1;

		ch=ReadToNextChar(tree_str,pos);

		if (ch=='[') {
			if (sscanf(&tree_str.at(++(*pos)), "%lf", &param)!=1) {
				sprintf(treeErrorMsg, "Unable to read branch parameter");
				return NULL;
			}
			ReadUntil(tree_str, pos, ']', "Close square bracket");
		} else { (*pos)--; }
	} else {
		if (tree->lengths==1) {
			sprintf(treeErrorMsg, "Some branches don't have branch lengths");
			return NULL;
		} else 
			tree->lengths=0;
	
		len=0.0;
		(*pos)--;
	}

	node->length0=len;
	node->param=param;
	
	return node;
}	

void TTree::ReadTree(string tree_str, int *pos, TTree *tree, bool first_partition)
{
	TNode *P;
	size_t found;

	treeError=0;
	tree->numNodes=0;
	tree->numTips=0;
	tree->rooted=1;
	tree->lengths=-1;

    if (tree_str.at(*pos) !='(' || (tree->root=ReadNode(tree_str, pos, tree, 0, first_partition))==NULL) {
		fprintf(stderr, "Error reading tree: %s.\n", treeErrorMsg);
		exit(EXIT_FAILURE);
    }

    found = tree_str.find(",);",++(*pos));

    if (found != tree_str.npos && tree_str.at(*pos) ==',') {		
		tree->rooted=0;
		if ((tree->root->branch0=ReadBranch(tree_str, pos, tree, first_partition))==NULL) {
	    	fprintf(stderr, "Error reading tree: %s.\n", treeErrorMsg);
	    	exit(EXIT_FAILURE);
		}
		tree->root->branch0->branch0=tree->root;
		tree->root->length0=tree->root->branch0->length0;
    }

    tree->totalLength=0.0;

    if (tree->rooted) {
		P=tree->root;
		while (P!=NULL) {
	    	tree->totalLength+=P->length0;
	    	P=P->branch1;
		}
    }
    
    for (list<inClade*>::iterator sit = treeEnv.begin(); sit != treeEnv.end(); sit++) {
    	for (int i = 0; i < names.size(); i++) {
    		if (((names.at(i)).compare((*sit)->clade_name) == 0)) {
				cerr << "Error parsing tree: Clade name '" << names.at(i) << "' is the name of a taxon." << endl;
				exit(EXIT_FAILURE);
			}
    	}
    }
}

void TTree::report_clades() 
{
	cerr << "Clades: " << endl;
	for (list<inClade*>::iterator it = treeEnv.begin(); it != treeEnv.end(); it++) {
		cerr << (*it)->clade_name << " reporting for duty.";
		if (!(*it)->environment_name.empty())
			cerr << " But you can call me " << (*it)->environment_name << ".";
		cerr << endl;
	}
}

char ReadToNextChar(string tree_str, int *pos)
{
	int found;	

	found = tree_str.find_first_not_of(" \t\n",++(*pos));
	*pos = found;

	return tree_str.at(*pos);
}

void ReadUntil(string tree_str, int *pos, char stopChar, char *what)
{
	size_t found;
	string delims = "(,:);";
	delims += stopChar;

	found = tree_str.find_first_of(delims,++(*pos));
	*pos = found;

	if (found == tree_str.npos || tree_str.at(*pos)!=stopChar) {
		sprintf(treeErrorMsg, "%s missing", what);
		treeError=1;
	}
}
